import copy

import numpy as np
import scipy.ndimage as ndi
import scipy.fft as fft
import pandas as pd

from copy import deepcopy
from scipy.special import factorial


def pool_voxel_grids(x1, x2, pooling_method):

    if x1 is None:
        return copy.deepcopy(x2)

    elif pooling_method == "max":
        # Perform max pooling by selecting the maximum intensity of each voxel.
        return np.maximum(x1, x2)

    elif pooling_method == "min":
        # Perform min pooling by selecting the minimum intensity of each voxel.
        return np.minimum(x1, x2)

    elif pooling_method in ["mean", "sum"]:
        # Perform mean / sum pooling by summing the intensities of each voxel.
        return np.add(x1, x2)

    else:
        raise ValueError(f"Unknown pooling method encountered: {pooling_method}")


class SeparableFilterSet:
    def __init__(
            self,
            filter_x,
            filter_y,
            filter_z=None,
            pre_filter_x=None,
            pre_filter_y=None,
            pre_filter_z=None
    ):
        self.x = filter_x
        self.y = filter_y
        self.z = filter_z

        self.pr_x = pre_filter_x
        self.pr_y = pre_filter_y
        self.pr_z = pre_filter_z

        # Extend even-sized filters.
        for attr in ["x", "y", "z", "pr_x", "pr_y", "pr_z"]:
            if self.__dict__[attr] is not None:

                # Check if the kernel is even or odd.
                if len(self.__dict__[attr]) % 2 == 0:
                    self.__dict__[attr] = np.append(self.__dict__[attr], 0.0)

    def permute_filters(self, rotational_invariance=True, require_pre_filter=False, as_filter_table=False):

        if require_pre_filter:
            if self.pr_x is None or self.pr_y is None:
                raise ValueError("The pre-filter should be set for all dimensions.")

            if self.z is not None and self.pr_z is None:
                raise ValueError("The pre-filter should have a component in the z-direction.")

            elif self.z is None and self.pr_z is not None:
                raise ValueError("The pre-filter should not have a component in the z-direction.")

        # Return an encapsulated version of the object.
        if not rotational_invariance:
            return [self]

        permuted_filters = []

        # Initiate filter strings
        g_x = "gx"
        g_y = "gy"
        g_z = "gz"
        jg_x = "jgx"
        jg_y = "jgy"
        jg_z = "jgz"

        # Test if x and y filters are the same.
        if np.array_equiv(self.x, self.y):
            g_y = g_x
            jg_y = jg_x

        # Test if x and z filters are the same
        if self.z is not None:
            if np.array_equiv(self.x, self.z):
                g_z = g_x
                jg_z = jg_x

        # Test if y and z filters are the same
        if self.z is not None:
            if np.array_equiv(self.y, self.z):
                g_z = g_y
                jg_z = jg_y

        # Test if the x-filter is symmetric.
        if np.array_equiv(self.x, np.flip(self.x)):
            jg_x = g_x

        # Test if the y-filter is symmetric.
        if np.array_equiv(self.y, np.flip(self.y)):
            jg_y = g_y

        # Test if the y-filter is symmetric.
        if self.z is not None:
            if np.array_equiv(self.z, np.flip(self.z)):
                jg_z = g_z

        if self.z is None:
            # 2D right-hand permutations
            permuted_filters += [{"x": g_x, "y": g_y}]
            permuted_filters += [{"x": jg_y, "y": g_x}]
            permuted_filters += [{"x": jg_x, "y": jg_y}]
            permuted_filters += [{"x": g_y, "y": jg_x}]

        else:
            # 3D right-hand permutations
            permuted_filters += [{"x": g_x, "y": g_y, "z": g_z}]
            permuted_filters += [{"x": jg_z, "y": g_y, "z": g_x}]
            permuted_filters += [{"x": jg_x, "y": g_y, "z": jg_z}]
            permuted_filters += [{"x": g_z, "y": g_y, "z": jg_x}]

            permuted_filters += [{"x": g_y, "y": g_z, "z": g_x}]
            permuted_filters += [{"x": g_y, "y": jg_z, "z": jg_x}]
            permuted_filters += [{"x": g_y, "y": jg_x, "z": g_z}]
            permuted_filters += [{"x": jg_x, "y": jg_y, "z": g_z}]

            permuted_filters += [{"x": jg_y, "y": g_x, "z": g_z}]
            permuted_filters += [{"x": jg_z, "y": jg_x, "z": g_y}]
            permuted_filters += [{"x": jg_z, "y": jg_y, "z": jg_x}]
            permuted_filters += [{"x": jg_z, "y": g_x, "z": jg_y}]

            permuted_filters += [{"x": jg_y, "y": jg_x, "z": jg_z}]
            permuted_filters += [{"x": g_x, "y": jg_y, "z": jg_z}]
            permuted_filters += [{"x": g_y, "y": g_x, "z": jg_z}]
            permuted_filters += [{"x": g_z, "y": jg_x, "z": jg_y}]

            permuted_filters += [{"x": g_z, "y": jg_y, "z": g_x}]
            permuted_filters += [{"x": g_z, "y": g_x, "z": g_y}]
            permuted_filters += [{"x": jg_x, "y": g_z, "z": g_y}]
            permuted_filters += [{"x": jg_y, "y": g_z, "z": jg_x}]

            permuted_filters += [{"x": g_x, "y": g_z, "z": jg_y}]
            permuted_filters += [{"x": jg_x, "y": jg_z, "z": jg_y}]
            permuted_filters += [{"x": jg_y, "y": jg_z, "z": g_x}]
            permuted_filters += [{"x": g_x, "y": jg_z, "z": g_y}]

        # Combine filters into a table.
        permuted_filters = pd.DataFrame(permuted_filters)

        if require_pre_filter:
            # Create a pre-filter to derive a table with filter orientations.
            pre_filter_set = SeparableFilterSet(filter_x=self.pr_x,
                                                filter_y=self.pr_y,
                                                filter_z=self.pr_z)

            permuted_pre_filters = pre_filter_set.permute_filters(rotational_invariance=rotational_invariance,
                                                                  as_filter_table=True)

            # Update the columns names
            permuted_pre_filters.rename(columns={"x": "pr_x",
                                                 "y": "pr_y",
                                                 "z": "pr_z"},
                                        inplace=True)

            # Join with the permuted_filters table.
            permuted_filters = pd.concat([permuted_pre_filters, permuted_filters], axis=1)

        if as_filter_table:
            return permuted_filters

        # Remove duplicates.
        permuted_filters = permuted_filters.drop_duplicates(ignore_index=True)

        filter_set_list = []
        for ii in range(len(permuted_filters)):
            permuted_filter_set = permuted_filters.loc[ii, :]

            filter_obj = deepcopy(self)

            if require_pre_filter:
                if self.z is None:
                    filter_set_list += [SeparableFilterSet(
                        filter_x=filter_obj._translate_filter(permuted_filter_set.x),
                        filter_y=filter_obj._translate_filter(permuted_filter_set.y),
                        pre_filter_x=filter_obj._translate_filter(permuted_filter_set.pr_x, True),
                        pre_filter_y=filter_obj._translate_filter(permuted_filter_set.pr_y, True)
                    )]

                else:
                    filter_set_list += [SeparableFilterSet(
                        filter_x=filter_obj._translate_filter(permuted_filter_set.x),
                        filter_y=filter_obj._translate_filter(permuted_filter_set.y),
                        filter_z=filter_obj._translate_filter(permuted_filter_set.z),
                        pre_filter_x=filter_obj._translate_filter(permuted_filter_set.pr_x, True),
                        pre_filter_y=filter_obj._translate_filter(permuted_filter_set.pr_y, True),
                        pre_filter_z=filter_obj._translate_filter(permuted_filter_set.pr_z, True)
                    )]

            else:
                if self.z is None:
                    filter_set_list += [SeparableFilterSet(
                        filter_x=filter_obj._translate_filter(permuted_filter_set.x),
                        filter_y=filter_obj._translate_filter(permuted_filter_set.y)
                    )]

                else:
                    filter_set_list += [SeparableFilterSet(
                        filter_x=filter_obj._translate_filter(permuted_filter_set.x),
                        filter_y=filter_obj._translate_filter(permuted_filter_set.y),
                        filter_z=filter_obj._translate_filter(permuted_filter_set.z)
                    )]

        return filter_set_list

    def _translate_filter(self, filter_symbol, use_pre_filter=False):

        if filter_symbol == "gx":
            if use_pre_filter:
                return self.pr_x
            else:
                return self.x

        elif filter_symbol == "gy":
            if use_pre_filter:
                return self.pr_y
            else:
                return self.y

        elif filter_symbol == "gz":
            if use_pre_filter:
                return self.pr_z
            else:
                return self.z

        elif filter_symbol == "jgx":
            if use_pre_filter:
                return np.flip(self.pr_x)
            else:
                return np.flip(self.x)

        elif filter_symbol == "jgy":
            if use_pre_filter:
                return np.flip(self.pr_y)
            else:
                return np.flip(self.y)

        elif filter_symbol == "jgz":
            if use_pre_filter:
                return np.flip(self.pr_z)
            else:
                return np.flip(self.z)

        else:
            raise ValueError(f"Encountered unrecognised filter symbol: {filter_symbol}")

    def decompose_filter(self, method="a_trous"):

        if method == "a_trous":
            # Add in 0s for the à trous algorithm

            # Iterate over filters.
            for attr in ["x", "y", "z", "pr_x", "pr_y", "pr_z"]:
                if self.__dict__[attr] is not None:
                    # Strip zeros from tail and head.
                    # old_filter_kernel = np.trim_zeros(deepcopy(self.__dict__[attr]))
                    old_filter_kernel = deepcopy(self.__dict__[attr])

                    # Create an array of zeros
                    new_filter_kernel = np.zeros(len(old_filter_kernel) * 2 - 1, dtype=float)

                    # Place the original filter constants at every second position. This creates a hole (0.0) between
                    # each of the filter constants.
                    new_filter_kernel[::2] = old_filter_kernel

                    # Update the attribute.
                    self.__dict__[attr] = new_filter_kernel

        else:
            raise ValueError(f"Unknown filter decomposition method: {method}")

    def convolve(self, voxel_grid, mode, use_pre_filter=False):

        # Ensure that we work from a local copy of voxel_grid to prevent updating it by reference.
        voxel_grid = deepcopy(voxel_grid)

        if use_pre_filter:
            if self.pr_x is None or self.pr_y is None or (self.z is not None and self.pr_z is None):
                raise ValueError("Pre-filter kernels are expected, but not found.")

            # Apply filter along the z-axis. Note that the voxel grid is stored with z, y, x indexing. Hence the
            # z-axis is the first axis, the y-axis the second, and the x-axis the third.
            if self.pr_z is not None:
                voxel_grid = ndi.convolve1d(voxel_grid, weights=self.pr_z, axis=0, mode=mode)

            # Apply filter along the y-axis.
            voxel_grid = ndi.convolve1d(voxel_grid, weights=self.pr_y, axis=1, mode=mode)

            # Apply filter along the x-axis.
            voxel_grid = ndi.convolve1d(voxel_grid, weights=self.pr_x, axis=2, mode=mode)

        else:
            # Apply filter along the z-axis. Note that the voxel grid is stored with z, y, x indexing. Hence the
            # z-axis is the first axis, the y-axis the second, and the x-axis the third.
            if self.z is not None:
                voxel_grid = ndi.convolve1d(voxel_grid, weights=self.z, axis=0, mode=mode)

            # Apply filter along the y-axis.
            voxel_grid = ndi.convolve1d(voxel_grid, weights=self.y, axis=1, mode=mode)

            # Apply filter along the x-axis.
            voxel_grid = ndi.convolve1d(voxel_grid, weights=self.x, axis=2, mode=mode)

        return voxel_grid


class FilterSet:
    def __init__(
            self,
            filter_set: np.ndarray,
            transformed=False,
            pad_image=True,
            riesz_order: None | list[int] = None,
            riesz_steered: bool = False,
            riesz_sigma: None | float = None
    ):

        self.filter_set = filter_set
        self.transformed = transformed
        self.pad_image = pad_image
        self.riesz_order = riesz_order
        self.riesz_steered = riesz_steered
        self.riesz_sigma = riesz_sigma

    def _get_coordinate_grid(self, axis=None):

        # Determine the grid center.
        grid_center = (np.array(self.filter_set.shape, dtype=float) - 1.0) / 2.0

        # Determine distance from center
        coordinate_grid = list(np.indices(self.filter_set.shape, sparse=True))
        coordinate_grid = [(coordinate_grid[ii] - center_pos) / center_pos for ii, center_pos in enumerate(grid_center)]

        # Broadcast to filter size shape.
        if axis is not None:
            coordinate_grid = np.broadcast_to(coordinate_grid[axis], self.filter_set.shape)

        return coordinate_grid

    def _get_distance_grid(self):

        # Generate a coordinate grid.
        coordinate_grid = self._get_coordinate_grid()

        # Compute the distances in the grid.
        distance_grid = np.sqrt(np.sum(
            np.power(np.meshgrid(coordinate_grid[0], coordinate_grid[1], coordinate_grid[2]), 2.0), axis=0)
        )

        return distance_grid

    def _pad_image(
            self,
            voxel_grid: np.ndarray,
            mode: str,
            axis: None | int = None
    ):

        # Modes in scipy and numpy are defined differently.
        if mode == "reflect":
            mode = "symmetric"
        elif mode == "symmetric":
            mode = "reflect"
        elif mode == "nearest":
            mode = "edge"
        elif mode == "mirror":
            mode = "reflect"

        # Ensure that we work from a local copy of voxel_grid to prevent updating it by reference.
        voxel_grid = deepcopy(voxel_grid)

        # Determine the original shape
        original_shape = voxel_grid.shape

        # Pad original image with half the kernel size on axes other than axis.
        if self.pad_image:
            pad_size = np.floor(np.array(self.filter_set.shape) / 2.0)
        else:
            pad_size = np.zeros(len(self.filter_set.shape))

        # Determine pad widths.
        pad_width = []
        original_offset = []
        axis_id = 0

        for current_axis in range(3):
            if current_axis != axis:
                # Select current axis.
                current_pad_size = int(pad_size[axis_id])

                # Add to elements.
                pad_width.append((current_pad_size, current_pad_size))
                original_offset.append(current_pad_size)

                # Update axis_id to skip to next element.
                axis_id += 1

            else:
                pad_width.append((0, 0))
                original_offset.append(0)

        # Set padding
        voxel_grid = np.pad(voxel_grid, pad_width=pad_width, mode=mode)

        return voxel_grid, original_shape, original_offset

    def _transform_filter(self, filter_shape, transform_method="interpolate"):

        if transform_method not in ["zero_pad", "interpolate"]:
            raise ValueError(
                f"The transform_method argument expects \"zero_pad\" or \"interpolate\". Found: {transform_method}"
            )

        if self.transformed and not np.equal(self.filter_set.shape, filter_shape).all() and transform_method == \
                "zero_pad":
            # Zero-padding takes place in the spatial domain. We therefore have to inverse transform the filter.
            self.filter_set = fft.ifftn(self.filter_set)

            # Transform back to the Fourier domain
            self.filter_set = fft.fftn(self.filter_set,
                                       filter_shape)

        elif self.transformed and not np.equal(self.filter_set.shape, filter_shape).all() and transform_method == \
                "interpolate":
            # Find zoom factor.
            zoom_factor = np.divide(filter_shape, self.filter_set.shape)

            # Make sure to zoom in the centric filter view.
            self.filter_set = fft.ifftshift(ndi.zoom(fft.fftshift(self.filter_set),
                                                     zoom=zoom_factor))

        elif not self.transformed:
            # Transform to the Fourier domain
            self.filter_set = fft.fftn(self.filter_set,
                                       filter_shape)

            self.transformed = True

    def _riesz_transform(self):

        # Skip if no Riesz transformation order have been set.
        if self.riesz_order is None:
            return

        # Skip if all Riesz transformation orders are 0.
        if not np.any(np.array(self.riesz_order) > 0):
            return

        # Check if the number of dimensions match.
        if len(self.riesz_order) is not np.ndim(self.filter_set):
            raise ValueError(f"The number of transformation orders ({len(self.riesz_order)}) does not match the filter "
                             f"dimension ({np.ndim(self.filter_set)}).")

        # Determine the order sum (L).
        order_sum = np.sum(np.array(self.riesz_order))

        # Compute the pre-factor, see equation 4.11.
        prefactor = (-1.0j) ** order_sum * np.sqrt(factorial(order_sum) / np.prod(factorial(np.array(
            self.riesz_order))))

        # Distance grid, see equation 4.11.
        distance_grid = np.power(self._get_distance_grid(), order_sum)

        # Set up gradient grid.
        gradient_grid = np.ones(self.filter_set.shape, dtype=np.cdouble)

        # Iterate over transformation orders.
        for ii, riesz_transform_order in enumerate(self.riesz_order):
            # Skip if no transformation is required.
            if riesz_transform_order == 0:
                continue

            # Update the gradient grid.
            gradient_grid = np.multiply(gradient_grid,
                                        np.power(self._get_coordinate_grid(axis=ii), riesz_transform_order))

        # Divide by l2-norm.
        gradient_grid = prefactor * np.divide(gradient_grid, distance_grid)

        # Update filter by taking Hadamard product.
        self.filter_set = fft.ifftshift(np.multiply(fft.fftshift(self.filter_set), gradient_grid))

    @staticmethod
    def _return_response(voxel_grid, response, original_offset, original_shape):
        # Compute response map.
        if response in ["modulus", "abs", "magnitude"]:
            voxel_grid = np.abs(voxel_grid)

        elif response in ["angle", "phase", "argument"]:
            voxel_grid = np.angle(voxel_grid)

        elif response in ["real"]:
            voxel_grid = np.real(voxel_grid)

        elif response in ["imaginary"]:
            voxel_grid = np.imag(voxel_grid)

        else:
            raise ValueError(f"The response argument should be \"modulus\", \"abs\", \"magnitude\", \"angle\", "
                             f"\"phase\", \"argument\", \"real\" or \"imaginary\". Found: {response}")

        # Crop image to original size.
        voxel_grid = voxel_grid[2 * original_offset[0]:2 * original_offset[0] + original_shape[0],
                                2 * original_offset[1]:2 * original_offset[1] + original_shape[1],
                                2 * original_offset[2]:2 * original_offset[2] + original_shape[2]]

        return voxel_grid


class FilterSet2D(FilterSet):

    def convolve(self,
                 voxel_grid: np.ndarray,
                 mode,
                 response,
                 axis=0):

        # Pad the image prior to convolution so that the valid convolution spans the image.
        voxel_grid, original_shape, original_offset = self._pad_image(voxel_grid, mode=mode, axis=axis)

        # Determine the filter output shape.
        filter_output_shape = [dim_size for current_axis, dim_size in enumerate(voxel_grid.shape) if
                               current_axis != axis]

        # Compute the fft of the filter, with output shape of the image.
        self._transform_filter(filter_shape=filter_output_shape)

        # Riesz transform the filter.
        self._riesz_transform()

        # Iterate over slices, compute fft, multiply with filter, and compute inverse fourier transform.
        voxel_grid = np.stack([self._convolve(voxel_grid=np.squeeze(current_grid, axis=axis))
                               for current_grid in np.split(voxel_grid, voxel_grid.shape[axis], axis=axis)],
                              axis=axis)

        return self._return_response(voxel_grid=voxel_grid,
                                     response=response,
                                     original_offset=original_offset,
                                     original_shape=original_shape)

    def _convolve(self,
                  voxel_grid: np.ndarray):

        # Compute FFT of the slice.
        f_voxel = fft.fft2(voxel_grid)

        if not self.transformed:
            raise ValueError("Filter should have been transformed to the Fourier domain.")

        # Multiply with the filter and return the inverse fourier transform of the hadamard product.
        return fft.ifft2(f_voxel * self.filter_set)


class FilterSet3D(FilterSet):

    def convolve(self,
                 voxel_grid: np.ndarray,
                 mode,
                 response):

        # Pad the image prior to convolution so that the valid convolution spans the image.
        voxel_grid, original_shape, original_offset = self._pad_image(voxel_grid, mode=mode)

        # Determine the filter output shape.
        filter_output_shape = [dim_size for current_axis, dim_size in enumerate(voxel_grid.shape)]

        # Compute the fft of the filter, with output shape of the image.
        self._transform_filter(filter_shape=filter_output_shape)

        # Riesz transform the filter.
        self._riesz_transform()

        # Iterate over slices, compute fft, multiply with filter, and compute inverse fourier transform.
        voxel_grid = self._convolve(voxel_grid)

        return self._return_response(voxel_grid=voxel_grid,
                                     response=response,
                                     original_offset=original_offset,
                                     original_shape=original_shape)

    def _convolve(self,
                  voxel_grid: np.ndarray):

        # Compute FFT of the slice
        f_voxel = fft.fftn(voxel_grid)

        if not self.transformed:
            raise ValueError("Filter should have been transformed to the Fourier domain.")

        # Multiply with the filter and return the inverse fourier transform of the hadamard product.
        return fft.ifftn(f_voxel * self.filter_set)



import warnings
import numpy as np
import datetime
from typing import Any

from mirp._data_import.dicom_file import ImageDicomFile
from mirp._data_import.dicom_multi_frame import ImageDicomMultiFrame
from mirp._data_import.utilities import parse_image_correction, convert_dicom_time, get_pydicom_meta_tag


class ImageDicomFilePT(ImageDicomFile):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def is_stackable(self, stack_images: str):
        return True

    def create(self):
        return self

    def load_data(
            self,
            pet_suv_conversion: str = "body_weight",
            **kwargs
    ):
        image_data = self.load_data_generic()

        conversion_possible = True
        # Get decay correction factor
        try:
            decay_factor = self._get_administration_decay_factor()
        except ValueError as err:
            warnings.warn(
                f"SUV cannot be computed as decay correction factor could not be determined. {str(err)}",
                UserWarning
            )
            conversion_possible = False
            decay_factor = 1.0

        # Get conversion factor to BQML
        try:
            bqml_factor = self._get_pet_unit_conversion_factor()
        except ValueError as err:
            if pet_suv_conversion != "none":
                warnings.warn(
                    f"SUV cannot be computed. BQML conversion factor could not be determined. {str(err)}",
                    UserWarning
                )
            conversion_possible = False
            bqml_factor = 1.0

        except NotImplementedError as err:
            if pet_suv_conversion != "none":
                warnings.warn(
                    f"SUV cannot be computed. BQML conversion factor could not be determined. {str(err)}",
                    UserWarning
                )
            conversion_possible = False
            bqml_factor = 1.0

        # Get SUV conversion factor and update the object_metadata attribute.
        if conversion_possible:
            suv_factor = self._get_suv_conversion_factor(new_suv_type=pet_suv_conversion)
            self.object_metadata.update(dict([("suv_type", pet_suv_conversion)]))

        else:
            suv_factor = 1.0
            self.object_metadata.update(dict([("suv_type", "none")]))

        # Update image_data
        image_data *= decay_factor * bqml_factor * suv_factor

        # Set image data.
        self.image_data = image_data

    def export_metadata(self, self_only=False, **kwargs) -> None | dict[str, Any]:
        if not self_only:
            metadata = super().export_metadata()
        else:
            metadata = {}

        self.load_metadata()

        dcm_meta_data = []

        # Time of flight information (0018,9755)
        time_of_flight = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9755),
            tag_type="str"
        )
        if time_of_flight is not None:
            dcm_meta_data += [("time_of_flight", time_of_flight)]

        # Radiopharmaceutical
        if get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x0016), tag_type=None, test_tag=True):
            radiopharmaceutical = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata[0x0054, 0x0016][0], tag=(0x0018, 0x0031), tag_type="str")
        else:
            radiopharmaceutical = None
        if radiopharmaceutical is not None:
            dcm_meta_data += [("radiopharmaceutical", radiopharmaceutical)]

        # Uptake time - acquisition start
        acquisition_ref_time = convert_dicom_time(
            date_str=get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0022), tag_type="str"),
            time_str=get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0032), tag_type="str")
        )

        # Uptake time - administration (0018,1078) is the administration start DateTime
        if get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x0016), test_tag=True):
            radio_admin_ref_time = convert_dicom_time(
                datetime_str=get_pydicom_meta_tag(
                    dcm_seq=self.image_metadata[0x0054, 0x0016][0], tag=(0x0018, 0x1078), tag_type="str")
            )

            if radio_admin_ref_time is None:
                # If unsuccessful, attempt determining administration time from (0x0018, 0x1072), which is the
                # administration start time.
                radio_admin_ref_time = convert_dicom_time(
                    date_str=get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0008, 0x0022), tag_type="str"),
                    time_str=get_pydicom_meta_tag(
                        dcm_seq=self.image_metadata[0x0054, 0x0016][0], tag=(0x0018, 0x1072), tag_type="str"
                    )
                )
        else:
            radio_admin_ref_time = None

        if radio_admin_ref_time is None:
            # If neither (0x0018, 0x1078) or (0x0018, 0x1072) are present, attempt to read private tags.
            # GE tags - note that due to anonymisation, acquisition time may be different then reported.
            acquisition_ref_time = convert_dicom_time(get_pydicom_meta_tag(
                dcm_seq=self.image_metadata, tag=(0x0009, 0x100d), tag_type="str"))
            radio_admin_ref_time = convert_dicom_time(get_pydicom_meta_tag(
                dcm_seq=self.image_metadata, tag=(0x0009, 0x103b), tag_type="str"))

        if radio_admin_ref_time is not None and acquisition_ref_time is not None:

            day_diff = abs(radio_admin_ref_time - acquisition_ref_time).days
            if day_diff > 1:
                # Correct for de-identification mistakes (i.e. administration time was de-identified correctly,
                # but acquisition time not). We do not expect that the difference between the two is more than a
                # day, or even more than a few hours at most.
                if radio_admin_ref_time > acquisition_ref_time:
                    radio_admin_ref_time -= datetime.timedelta(days=day_diff)
                else:
                    radio_admin_ref_time += datetime.timedelta(days=day_diff)

            if radio_admin_ref_time > acquisition_ref_time:
                # Correct for overnight
                radio_admin_ref_time -= datetime.timedelta(days=1)

            # Calculate uptake time in minutes
            uptake_time = ((acquisition_ref_time - radio_admin_ref_time).seconds / 60.0)
        else:
            uptake_time = None

        if uptake_time is not None:
            dcm_meta_data += [("uptake_time", uptake_time)]

        # Reconstruction method
        reconstruction_method = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0054, 0x1103),
            tag_type="str"
        )
        if reconstruction_method is not None:
            dcm_meta_data += [("reconstruction_method", reconstruction_method)]

        # Convolution kernel
        kernel = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1210),
            tag_type="str"
        )
        if kernel is not None:
            dcm_meta_data += [("kernel", kernel)]

        # Reconstruction type
        recon_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9756),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9749)
        )
        if recon_type is not None:
            dcm_meta_data += [("reconstruction_type", recon_type)]

        # Reconstruction algorithm
        recon_algorithm = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9315),
            tag_type="str",
            macro_dcm_seq=(0x0018, 0x9749)
        )
        if reconstruction_method is not None:
            dcm_meta_data += [("reconstruction_algorithm", recon_algorithm)]

        # Number of iterations
        n_iterations = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9739),
            tag_type="int",
            macro_dcm_seq=(0x0018, 0x9749)
        )
        if n_iterations is not None:
            dcm_meta_data += [("n_iterations", n_iterations)]

        # Number of subsets
        n_subsets = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9740),
            tag_type="int",
            macro_dcm_seq=(0x0018, 0x9749)
        )
        if n_subsets is not None:
            dcm_meta_data += [("n_subsets", n_subsets)]

        # Frame duration (converted from milliseconds to seconds)
        frame_duration = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x1242),
            tag_type="float"
        )
        if frame_duration is not None:
            dcm_meta_data += [("frame_duration", frame_duration / 1000.0)]

        # Image corrections
        image_corrections = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0028, 0x0051),
            tag_type="str"
        )
        if image_corrections is not None:
            dcm_meta_data += [("image_corrections", image_corrections)]

        # Attenuation corrected ATTN (0018,9759)
        attenuation_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9759),
            correction_abbr="ATTN"
        )
        if attenuation_corrected is not None:
            dcm_meta_data += [("attenuation_corrected", attenuation_corrected)]

        # Attenuation correction method
        attenuation_correction_method = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0054, 0x1101),
            tag_type="str"
        )
        if attenuation_corrected is not None:
            dcm_meta_data += [("attenuation_correction_method", attenuation_correction_method)]

        # Scatter corrected SCAT (0018,9760)
        scatter_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9760),
            correction_abbr="SCAT"
        )
        if scatter_corrected is not None:
            dcm_meta_data += [("scatter_corrected", scatter_corrected)]

        # Scatter correction method
        scatter_correction_method = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0054, 0x1105),
            tag_type="str"
        )
        if scatter_correction_method is not None:
            dcm_meta_data += [("scatter_correction_method", scatter_correction_method)]

        # Randoms corrected RAN (0018,9765)
        randoms_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9765),
            correction_abbr="RAN"
        )
        if randoms_corrected is not None:
            dcm_meta_data += [("randoms_corrected", randoms_corrected)]

        # Randoms correction method
        random_correction_method = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0054, 0x1100),
            tag_type="str"
        )
        if random_correction_method is not None:
            dcm_meta_data += [("random_correction_method", random_correction_method)]

        # Decay corrected DECY (0018,9758)
        decay_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9758),
            correction_abbr="DECY"
        )
        if decay_corrected is not None:
            dcm_meta_data += [("decay_corrected", decay_corrected)]

        # Dead time corrected DTIM (0018,9761)
        dead_time_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9761),
            correction_abbr="DTIM"
        )
        if dead_time_corrected is not None:
            dcm_meta_data += [("dead_time_corrected", dead_time_corrected)]

        # Gantry motion corrected MOTN (0018,9762)
        gantry_motion_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9762),
            correction_abbr="MOTN"
        )
        if gantry_motion_corrected is not None:
            dcm_meta_data += [("gantry_motion_corrected", gantry_motion_corrected)]

        # Patient motion corrected PMOT (0018,9763)
        patient_motion_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9763),
            correction_abbr="PMOT"
        )
        if patient_motion_corrected is not None:
            dcm_meta_data += [("patient_motion_corrected", patient_motion_corrected)]

        # Count loss normalisation corrected CLN (0018,9764)
        count_loss_norm_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9764),
            correction_abbr="CLN"
        )
        if count_loss_norm_corrected is not None:
            dcm_meta_data += [("count_loss_norm_corrected", count_loss_norm_corrected)]

        # Non-uniform radial sampling corrected RADL (0018,9766)
        radl_corrected = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9766),
            correction_abbr="RADL"
        )
        if radl_corrected is not None:
            dcm_meta_data += [("radl_corrected", radl_corrected)]

        # Sensitivity calibrated DCAL (0018,9767)
        sensitivity_calibrated = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9767),
            correction_abbr="DCAL"
        )
        if sensitivity_calibrated is not None:
            dcm_meta_data += [("sensitivity_calibrated", sensitivity_calibrated)]

        # Detector normalisation correction NORM (0018,9768)
        detector_normalisation = parse_image_correction(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9768),
            correction_abbr="NORM"
        )
        if detector_normalisation is not None:
            dcm_meta_data += [("detector_normalisation", detector_normalisation)]

        metadata.update(dict(dcm_meta_data))
        return metadata

    def _get_tracer_administration_time(self) -> datetime.datetime:
        self.load_metadata()

        # Set initial value of tracer administration reference time.
        admin_ref_time = None

        # Administration time should come from the Radiopharmaceutical Information Sequence (0x0054, 0x0016).
        has_sequence = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x0016), test_tag=True)

        # Prefer Radiopharmaceutical Start DateTime (0x0018, 0x1078)
        if has_sequence and admin_ref_time is None:
            admin_ref_time = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata[0x0054, 0x0016][0],
                tag=(0x0018, 0x1078),
                tag_type="str"
            )
            admin_ref_time = convert_dicom_time(datetime_str=admin_ref_time)

        # Fallback to Radiopharmaceutical Start Time (0x0018, 0x1072)
        if has_sequence and admin_ref_time is None:
            admin_ref_time = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata[0x0054, 0x0016][0],
                tag=(0x0018, 0x1072),
                tag_type="str"
            )

            if admin_ref_time is not None:
                # Infer start date.
                acquisition_start_time = self._get_acquisition_start_time()
                admin_ref_time = datetime.datetime(
                    year=acquisition_start_time.year,
                    month=acquisition_start_time.month,
                    day=acquisition_start_time.day,
                    hour=int(admin_ref_time[0:2]),
                    minute=int(admin_ref_time[2:4]),
                    second=int(admin_ref_time[4:6]),
                    microsecond=0 if len(admin_ref_time) <= 6 else int(round(float(admin_ref_time[6:]) * 1000))
                )

                # Correct for overnight recordings.
                if admin_ref_time > acquisition_start_time:
                    admin_ref_time -= datetime.timedelta(days=(acquisition_start_time - admin_ref_time).days)

        #  Fall back to Private GE Radiopharmaceutical Start DateTime.
        if admin_ref_time is None:
            admin_ref_time = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata,
                tag=(0x0009, 0x103b),
                tag_type="str"
            )
            admin_ref_time = convert_dicom_time(datetime_str=admin_ref_time)

        # Final check.
        if admin_ref_time is None:
            raise ValueError(
                f"Radiopharmaceutical start time cannot be determined from DICOM metadata. [{self.describe_self()}]"
            )

        return admin_ref_time

    def _get_administration_decay_factor(self) -> float:
        self.load_metadata()

        # Type of decay correction that is used
        decay_correction = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0054, 0x1102),
            tag_type="str",
            default="NONE"
        )
        if decay_correction == "ADMIN":
            return 1.0

        elif decay_correction not in ["NONE", "START"]:
            raise ValueError(
                f"Decay correction DICOM tag was not recognised: {decay_correction}. One of ",
                f"NONE, START or ADMIN was expected. [{self.describe_self()}]"
            )

        # Get acquisition start time and tracer administration time.
        acquisition_start_time = self._get_acquisition_start_time()
        tracer_administration_time = self._get_tracer_administration_time()

        # Get frame duration in seconds.
        frame_duration = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0018, 0x1242), tag_type="float")
        if frame_duration is None:
            raise ValueError(f"Frame duration cannot be determined from DICOM metadata. [{self.describe_self()}]")
        frame_duration /= 1000.0  # From milliseconds to seconds.

        # Radionuclide total dose and radionuclide half-life
        half_life = None
        if get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x0016), test_tag=True):
            half_life = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata[0x0054, 0x0016][0],
                tag=(0x0018, 0x1075),
                tag_type="float"
            )

        if half_life is None:
            raise ValueError(
                f"Radionuclide half-life (0x0018, 0x1075) was missing in the Radiopharmaceutical "
                f"information sequence (0x0054, 0x0016). [{self.describe_self()}]"
            )

        # Decay constant.
        _lambda = np.log(2.0) / half_life

        # Process for different decay corrections
        if decay_correction == "NONE":
            time_to_acquisition_start = acquisition_start_time - tracer_administration_time
            time_to_acquisition_start = (
                    time_to_acquisition_start.days * 86400.0 +
                    time_to_acquisition_start.seconds +
                    time_to_acquisition_start.microseconds / 1000000.0
            )

            # Correct for decay during frame in addition to decay between administration and acquisition start.
            decay_factor = (
                frame_duration * _lambda * np.exp(_lambda * time_to_acquisition_start) /
                (1.0 - np.exp(-_lambda * frame_duration))
            )

        elif decay_correction == "START":
            # Decay correction of pixel values for the period from pixel acquisition up to scan start
            # Additionally correct for decay between administration and acquisition start. Based on QIBA SUV
            # vendorneutral pseudocode.

            time_to_acquisition_start = acquisition_start_time - tracer_administration_time
            time_to_acquisition_start = (
                    time_to_acquisition_start.days * 86400.0 +
                    time_to_acquisition_start.seconds +
                    time_to_acquisition_start.microseconds / 1000000.0
            )

            # Correct for decay between administration and acquisition start.
            decay_factor = np.exp(_lambda * time_to_acquisition_start)

        else:
            raise ValueError(
                f"Decay correction DICOM tag was not recognised: {decay_correction}. One of NONE, START or ADMIN "
                f"was expected. [{self.describe_self()}]"
            )

        return decay_factor

    def _get_pet_unit_conversion_factor(self) -> float:
        """To compute SUV, PET units need to be converted to BQML."""
        self.load_metadata()

        pet_unit = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x1001), tag_type="str")
        if pet_unit is None:
            raise ValueError(f"PET Units (0x0054, 0x1001) was missing. [{self.describe_self()}]")

        if pet_unit in ["CNTS", "CPS"]:
            conversion_factor = self._pet_unit_cnt_to_bqml()
        elif pet_unit in ["BQML"]:
            conversion_factor = 1.0
        elif pet_unit in ["GML", "CM2ML"]:
            conversion_factor = 1.0
        else:
            raise NotImplementedError(
                f"Conversion factor for converting {pet_unit} to BQML is not implemented. [{self.describe_self()}]"
            )

        return conversion_factor

    def _pet_unit_cnt_to_bqml(self) -> float:
        self.load_metadata()

        pet_unit = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x1001), tag_type="str")

        # Read private tag.
        conversion_factor = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x7053, 0x1009), tag_type="float")

        # Use frame duration (if not CPS), and Radiopharmaceutical Volume to convert to Bq/ml.
        if conversion_factor is None:
            conversion_factor = 1.0
            if pet_unit == "CNTS":
                frame_duration = get_pydicom_meta_tag(
                    dcm_seq=self.image_metadata,
                    tag=(0x0018, 0x1242),
                    tag_type="float"
                )
                if frame_duration is None:
                    raise ValueError(
                        f"Frame duration cannot be determined from DICOM metadata. [{self.describe_self()}]"
                    )
                frame_duration /= 1000.0  # From milliseconds to seconds.
                conversion_factor = 1.0 / frame_duration

            # Radiopharmaceutical volume should come from the Radiopharmaceutical Information Sequence (0x0054, 0x0016).
            if get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x0016), test_tag=True):
                administered_volume = get_pydicom_meta_tag(
                    dcm_seq=self.image_metadata[0x0054, 0x0016][0],
                    tag=(0x0018, 0x1071),
                    tag_type="float"
                )
                if administered_volume is None:
                    raise ValueError(
                        f"Radiopharmaceutical volume cannot be determined from DICOM metadata. [{self.describe_self()}]"
                    )

                # Divide by administered volume (in cubic cm == milliliter)
                conversion_factor /= administered_volume

            else:
                raise ValueError(
                    f"Radiopharmaceutical Information Sequence (0x0054, 0x0016) is missing in DICOM metadata. "
                    f"[{self.describe_self()}]"
                )

        # Final check
        if conversion_factor is None:
            raise ValueError(
                f"Conversion factor for converting {pet_unit} to BQML could not be established. "
                f"[{self.describe_self()}]"
            )

        return conversion_factor

    def _get_suv_conversion_factor(self, new_suv_type: str) -> float:
        self.load_metadata()

        current_suv_type = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0054, 0x1006),
            tag_type="str"
        )

        # Set SUV type based on PET unit.
        if current_suv_type is None:
            pet_unit = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x1001), tag_type="str")
            if pet_unit is None:
                raise ValueError(f"PET Units (0x0054, 0x1001) was missing. [{self.describe_self()}]")

            if pet_unit == "GML":
                # If absent, and the Units (0054,1001) are GML, then the type of SUV shall be assumed to be BW.
                current_suv_type = "BW"
            elif pet_unit == "CM2ML":
                current_suv_type = "BSA"

        # If SUV type was not set, and cannot be inferred, assume that intensities do not represent SUV.
        if current_suv_type is None:
            current_suv_type = "none"

        # Convert DICOM SUV type to internal format.
        translation_table = dict([
            ("none", "none"),
            ("BW", "body_weight"),
            ("BSA", "body_surface_area"),
            ("LBM", "lean_body_mass_error"),
            ("LBMJAMES128", "lean_body_mass"),
            ("LBMJANMA", "lean_body_mass_bmi"),
            ("IBW", "ideal_body_weight")
        ])
        current_suv_type = translation_table[current_suv_type]

        if current_suv_type == new_suv_type:
            return 1.0

        # Convert back to BQML.
        revert_suv_factor = 1.0
        if current_suv_type != "none":
            revert_suv_factor = 1.0 / self._compute_suv_factor(suv_type=current_suv_type)

        suv_factor = 1.0
        if new_suv_type != "none":
            suv_factor = self._compute_suv_factor(suv_type=new_suv_type)

        return revert_suv_factor * suv_factor

    def _compute_suv_factor(self, suv_type: str) -> float:

        # No SUV -------------------------------------------------------------------------------------------------------
        if suv_type == "none":
            return 1.0

        # Require body weight and administered dose.
        patient_weight = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0010, 0x1030), tag_type="float")
        if patient_weight is None:
            raise ValueError(
                f"Patient weight (0x0010, 0x1030) was missing. SUV normalisation is not possible. "
                f"[{self.describe_self()}]"
            )
        elif patient_weight <= 0.0:
            raise ValueError(
                f"Patient weight (0x0010, 0x1030) was not positive ({patient_weight}). SUV normalisation is not "
                f"possible. [{self.describe_self()}]"
            )
        elif patient_weight >= 1000.0:
            # Weight is likely provide in grams, not kilograms. Convert to kg.
            patient_weight /= 1000.0

        # Administered dose should come from the Radiopharmaceutical Information Sequence (0x0054, 0x0016).
        administered_dose = None
        has_sequence = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0054, 0x0016), test_tag=True)
        if has_sequence and administered_dose is None:
            administered_dose = get_pydicom_meta_tag(
                dcm_seq=self.image_metadata[0x0054, 0x0016][0],
                tag=(0x0018, 0x1074),
                tag_type="float"
            )

        if administered_dose is None:
            raise ValueError(
                f"Radionuclide Total Dose (0x0018, 0x1074) was missing. SUV normalisation is not possible. "
                f"[{self.describe_self()}]"
            )
        elif administered_dose <= 0.0:
            raise ValueError(
                f"Radionuclide Total Dose (0x0018, 0x1074) was not positive ({administered_dose}). "
                f"SUV normalisation is not possible. [{self.describe_self()}]"
            )

        # Body weight-corrected SUV ------------------------------------------------------------------------------------
        if suv_type == "body_weight":
            return patient_weight * 1000.0 / administered_dose

        # Require patient height.
        patient_height = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0010, 0x1020), tag_type="float")
        if patient_height is None:
            raise ValueError(
                f"Patient Size (0x0010, 0x1020) was missing. SUV normalisation ({suv_type}) is not possible. "
                f"[{self.describe_self()}]"
            )
        elif patient_height <= 0.0:
            raise ValueError(
                f"Patient Size (0x0010, 0x1020) was not positive ({patient_height}). SUV normalisation ({suv_type}) "
                f"is not possible. [{self.describe_self()}]"
            )
        elif patient_height > 3.0:
            # Interpret patient height as cm and convert to meter.
            patient_height /= 100.0

        # Body surface area-corrected SUV ------------------------------------------------------------------------------
        if suv_type == "body_surface_area":
            # Kim et al. Journal of Nuclear Medicine. Volume 35, No. 1, January 1994. pp 164-167
            return 1000.0 * patient_weight ** 0.425 * (patient_height * 100.0) ** 0.725 * 0.007184 / administered_dose

        # Require patient biological sex.
        patient_biological_sex = get_pydicom_meta_tag(dcm_seq=self.image_metadata, tag=(0x0010, 0x0040), tag_type="str")
        if patient_biological_sex is None:
            patient_biological_sex = "O"
        if patient_biological_sex.lower() not in ["m", "f", "w", "o", "d", "u"]:
            raise ValueError(
                f"Patient Sex (0x0010, 0x0040) was not recognised ({patient_biological_sex}. SUV normalisation "
                f"({suv_type}) is not possible. [{self.describe_self()}]"
            )

        # Erroneous lean body mass-corrected SUV -----------------------------------------------------------------------
        if suv_type == "lean_body_mass_error":
            if patient_biological_sex.lower() == "m":
                norm_factor = 1.10 * patient_weight - 120.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
            elif patient_biological_sex.lower() in ["f", "w"]:
                norm_factor = 1.07 * patient_weight - 148.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
            elif patient_biological_sex.lower() in ["o", "d", "u"]:
                # Average for other, diverse or unknown -- not ideal, but better than throwing an error.
                norm_factor = (
                        1.10 * patient_weight - 120.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
                        + 1.07 * patient_weight - 148.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
                ) / 2.0
            else:
                raise ValueError("unreachable code")

            return norm_factor * 1000.0 / administered_dose

        # Lean body mass-corrected SUV ---------------------------------------------------------------------------------
        if suv_type == "lean_body_mass":
            if patient_biological_sex.lower() == "m":
                norm_factor = 1.10 * patient_weight - 128.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
            elif patient_biological_sex.lower() in ["f", "w"]:
                norm_factor = 1.07 * patient_weight - 148.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
            elif patient_biological_sex.lower() in ["o", "d", "u"]:
                # Average for other, diverse or unknown -- not ideal, but better than throwing an error.
                norm_factor = (
                        1.10 * patient_weight - 128.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
                        + 1.07 * patient_weight - 148.0 * (patient_weight ** 2.0 / (patient_height * 100.0) ** 2.0)
                ) / 2.0
            else:
                raise ValueError("unreachable code")

            return norm_factor * 1000.0 / administered_dose

        # Lean body mass (BMI)-corrected SUV ---------------------------------------------------------------------------
        if suv_type == "lean_body_mass_bmi":
            # Janmahasatian, Sarayut, et al. "Quantification of lean bodyweight." Clinical pharmacokinetics 44
            # (2005): 1051-1065.
            bmi = patient_weight / patient_height**2.0
            if patient_biological_sex.lower() in ["m"]:
                norm_factor = 9270.0 * patient_weight / (6680.0 + 216.0 * bmi)
            elif patient_biological_sex.lower() in ["f", "w"]:
                norm_factor = 9270.0 * patient_weight / (8780.0 + 244.0 * bmi)
            elif patient_biological_sex.lower() in ["o", "d", "u"]:
                # Average for other, diverse or unknown -- not ideal, but better than throwing an error.
                norm_factor = (
                    9270.0 * patient_weight / (6680.0 + 216.0 * bmi) + 9270.0 * patient_weight / (8780.0 + 244.0 * bmi)
                ) / 2.0
            else:
                raise ValueError("unreachable code")

            return norm_factor * 1000. / administered_dose

        # Ideal body weight (IBW)-corrected SUV ------------------------------------------------------------------------
        if suv_type == "ideal_body_weight":
            if patient_biological_sex.lower() in ["m"]:
                norm_factor = 48.0 + 1.06 * (patient_height * 100.0 - 152.0)
            elif patient_biological_sex.lower() in ["f", "w"]:
                norm_factor = 45.5 + 0.91 * (patient_height * 100.0 - 152.0)
            elif patient_biological_sex.lower() in ["o", "d", "u"]:
                # Average for other, diverse or unknown -- not ideal, but better than throwing an error.
                norm_factor = (
                        48.0 + 1.06 * (patient_height * 100.0 - 152.0) + 45.5 + 0.91 * (patient_height * 100.0 - 152.0)
                ) / 2.0

            else:
                raise ValueError("unreachable code")

            return norm_factor * 1000.0 / administered_dose


class ImageDicomFilePTMultiFrame(ImageDicomMultiFrame, ImageDicomFilePT):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


import ctypes
import os
import warnings
from ctypes.util import find_library

# Check if the ray package is available
RAY_AVAILABLE = True
try:
    import ray
except ImportError:
    RAY_AVAILABLE = False


def limit_inner_threads(n_threads: int = 1):
    # OpenBLAS-based multi-threading libraries
    try_paths = [
        '/opt/OpenBLAS/lib/libopenblas.so',
        '/lib/libopenblas.so',
        '/usr/lib/libopenblas.so.0',
        find_library('openblas')
    ]

    # openBLAS library
    openblas_lib = None
    for libpath in try_paths:
        try:
            openblas_lib = ctypes.cdll.LoadLibrary(libpath)
            break
        except (OSError, TypeError):
            continue

    if openblas_lib is not None:
        try:
            openblas_lib.openblas_set_num_threads(n_threads)
        except:
            pass

    # MKL library
    try:
        import mkl
        mkl.set_num_threads(n_threads)
    except:
        pass

    # Set OS variables.
    os.environ["OMP_NUM_THREADS"] = str(n_threads)
    os.environ["OPENBLAS_NUM_THREADS"] = str(n_threads)
    os.environ["MKL_NUM_THREADS"] = str(n_threads)
    os.environ["BLIS_NUM_THREADS"] = str(n_threads)
    os.environ["VECLIB_MAXIMUM_THREADS"] = str(n_threads)
    os.environ["NUMBA_NUM_THREADS"] = str(n_threads)
    os.environ["NUMEXPR_NUM_THREADS"] = str(n_threads)


def ray_remote_disabled(func):
    def placeholder_function(*args, **kwargs):
        pass
    return placeholder_function


def ray_is_available():
    return RAY_AVAILABLE


def ray_is_initialized():
    if RAY_AVAILABLE:
        return ray.is_initialized()
    else:
        return False


def ray_init(num_cpus):
    if RAY_AVAILABLE:
        try:
            ray.init(num_cpus=num_cpus)
        except OSError as err:
            warnings.warn(
                f"Ray instances could not be started for parallel processing. Switching to sequential processing. "
                f"{str(err)}",
                UserWarning
            )
            pass
    else:
        warnings.warn(
            "The ray package was not found. Switching to sequential processing.",
            UserWarning
        )


def ray_get(x):
    if RAY_AVAILABLE:
        return ray.get(x)
    else:
        raise ModuleNotFoundError(
            "The ray package was not found. No results can be obtained."
        )


def ray_shutdown():
    if RAY_AVAILABLE:
        ray.shutdown()


ray_remote = ray.remote if RAY_AVAILABLE else ray_remote_disabled



import warnings

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

from mirp._images.generic_image import GenericImage
from mirp._images.mask_image import MaskImage
from mirp._masks.base_mask import BaseMask


class InteractivePlot:  # pragma: no cover

    def __init__(
            self,
            image: GenericImage,
            mask: MaskImage | BaseMask | None = None,
            slice_id: int | None = None
    ):

        # Determine if a mask should be shown.
        show_mask = mask is not None and not mask.is_empty()

        if show_mask and isinstance(mask, BaseMask):
            mask = mask.roi
        if show_mask:
            show_mask = not mask.is_empty_mask()

        # Generate figure
        figure, axes = plt.subplots()
        self.axes = axes
        self.figure = figure

        # Attach connections
        self.scroll_cid = self.figure.canvas.mpl_connect("scroll_event", self.onscroll)
        self.close_cid = self.figure.canvas.mpl_connect("close_event", self.disconnect)

        self.image_data = image.get_voxel_grid()

        self.mask_data = None
        if show_mask:
            self.mask_data = mask.get_voxel_grid()

        self.n_slices, _, _ = image.image_dimension
        if slice_id is None:
            self.slice_index = int(np.floor(self.n_slices / 2.0))
        elif not isinstance(slice_id, int):
            raise TypeError(f"slice_id should be an integer (int). Found: {type(slice_id)}")
        elif slice_id < 1:
            self.slice_index = 0
            warnings.warn(f"slice_id cannot be smaller than 1. Found: {slice_id}", UserWarning)
        elif slice_id > self.n_slices:
            self.slice_index = self.n_slices - 1
            warnings.warn(
                f"slice_id cannot be greater than the number of slices ({self.n_slices}). Found: {slice_id}",
                UserWarning
            )
        else:
            self.slice_index = slice_id - 1

        # Set plotting options
        colour_map = image.get_colour_map()
        min_intensity = image.get_default_lowest_intensity() if image.get_default_lowest_intensity() is not None else (
            np.min(self.image_data))
        max_intensity = image.get_default_upper_intensity() if image.get_default_upper_intensity() is not None else (
            np.max(self.image_data))

        # Create plot.
        self.image_layer = self.axes.imshow(
            self.image_data[self.slice_index, :, :],
            vmin=min_intensity,
            vmax=max_intensity,
            cmap=colour_map
        )

        # Create mask.
        if show_mask:
            from importlib_metadata import version
            from packaging.version import Version

            # Define color map. The custom color map goes from transparent black to semi-transparent green and is
            # used as an overlay. Note that register_cmap is deprecated in version 3.9.0 of matplotlib
            if Version(version("matplotlib")) >= Version("3.9.0"):
                import matplotlib

                matplotlib.colormaps.register(
                    cmap=LinearSegmentedColormap(
                        "mask_cm",
                        {
                            'red': ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
                            'green': ((0.0, 0.0, 0.0), (1.0, 0.6, 0.6)),
                            'blue': ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
                            'alpha': ((0.0, 0.0, 0.0), (1.0, 0.4, 0.4))
                        }
                    ),
                    force=True
                )

            else:
                # Create map and register
                plt.register_cmap(
                    cmap=LinearSegmentedColormap(
                        "mask_cm",
                        {
                            'red': ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
                            'green': ((0.0, 0.0, 0.0), (1.0, 0.6, 0.6)),
                            'blue': ((0.0, 0.0, 0.0), (1.0, 0.0, 0.0)),
                            'alpha': ((0.0, 0.0, 0.0), (1.0, 0.4, 0.4))
                        }
                    ),
                    override_builtin=True
                )

            self.mask_layer = self.axes.imshow(
                self.mask_data[self.slice_index, :, :],
                vmin=0.0,
                vmax=1.0,
                cmap="mask_cm"
            )
        else:
            self.mask_layer = None

        self.update()

    def onscroll(self, event):
        # Update, but limit to range of available slices.
        if event.button == "up":
            self.slice_index += 1
            if self.slice_index >= self.n_slices:
                self.slice_index -= 1
        else:
            self.slice_index -= 1
            if self.slice_index < 0:
                self.slice_index += 1
        self.update()

    def update(self):
        self.axes.set_title(f"slice {self.slice_index + 1}")
        self.image_layer.set_data(self.image_data[self.slice_index, :, :])
        self.image_layer.axes.figure.canvas.draw()

        if self.mask_layer is not None:
            self.mask_layer.set_data(self.mask_data[self.slice_index, :, :])
            self.mask_layer.axes.figure.canvas.draw()

    def disconnect(self):
        self.figure.canvas.mpl_disconnect(self.scroll_cid)
        self.figure.canvas.mpl_disconnect(self.close_cid)


import copy
import os
import warnings
from typing import Any
from xml.etree import ElementTree as ElemTree

from mirp.settings.utilities import str2list, str2type, read_node


def import_data_settings(
    path: str,
    is_mask: bool
) -> list[dict[str, Any]]:
    if not os.path.exists(path):
        raise ValueError(f"The {path} data settings file does not exist. Please check spelling of the file path.")
    if not path.endswith(".xml"):
        raise ValueError(f"The {path} data settings file is not an xml file.")

    # Load xml
    tree = ElemTree.parse(path)
    root = tree.getroot()

    # Empty list for iteratively storing data objects
    data_arguments_list = []

    # Iterate over configurations
    for branch in root.findall("config"):
        # Set data arguments.
        data_arguments = []

        # Read data from xml file
        paths_branch = branch.find("paths")

        # Set main directory.
        if is_mask:
            mask = os.path.normpath(
                str2type(read_node(paths_branch, ["project_folder", "mask"]), "path"))
            data_arguments += [("image", mask)]
        else:
            image = os.path.normpath(
                str2type(read_node(paths_branch, ["project_folder", "image"]), "path"))
            data_arguments += [("image", image)]

        # Set sample name.
        sample_name = str2list(read_node(paths_branch, ["sample_name", "subject_include"]), "str")
        data_arguments += [("sample_name", sample_name)]

        # Deprecated items.
        if paths_branch.find("subject_exclude") is not None:
            # noinspection PyDeprecation
            _import_data_settings_deprecation_warning("subject_exclude")
        if paths_branch.find("write_folder") is not None:
            # noinspection PyDeprecation
            _import_data_settings_deprecation_warning("write_path")
        if paths_branch.find("provide_diagnostics") is not None:
            # noinspection PyDeprecation
            _import_data_settings_deprecation_warning("provide_diagnostics")
        if paths_branch.find("cohort") is not None:
            # noinspection PyDeprecation
            _import_data_settings_deprecation_warning("cohort")
       
        # Iterate over data branches
        for data_branch in branch.findall("data"):
            current_data_arguments = copy.deepcopy(data_arguments)

            if is_mask:
                mask_name = str2list(data_branch.find("mask_name"), "str")
                if mask_name is not None and len(mask_name) == 1:
                    mask_name = mask_name[0]
                current_data_arguments += [("image_name", mask_name)]

                mask_file_type = str2type(data_branch.find("mask_file_type"), "str")
                current_data_arguments += [("image_file_type", mask_file_type)]

                mask_sub_folder = str2type(read_node(data_branch, ["mask_sub_folder", "roi_folder"]), "path")
                current_data_arguments += [("image_sub_folder", mask_sub_folder)]

                mask_modality = str2type(data_branch.find("mask_modality"), "str")
                current_data_arguments += [("image_modality", mask_modality)]

                roi_name = str2list(read_node(data_branch, ["roi_name", "roi_names"]), "str")
                current_data_arguments += [("roi_name", roi_name)]

            else:
                image_name = str2list(read_node(data_branch, ["image_name", "image_filename_pattern"]), "str")
                if image_name is not None and len(image_name) == 1:
                    image_name = image_name[0]
                current_data_arguments += [("image_name", image_name)]

                image_file_type = str2type(data_branch.find("image_file_type"), "str")
                current_data_arguments += [("image_file_type", image_file_type)]

                image_sub_folder = str2type(read_node(data_branch, ["image_sub_folder", "image_folder"]), "path")
                current_data_arguments += [("image_sub_folder", image_sub_folder)]

                image_modality = str2type(read_node(data_branch, ["image_modality", "modality"]), "str")
                current_data_arguments += [("image_modality", image_modality)]

            # More deprecated items.
            if data_branch.find("registration_image_folder") is not None:
                # noinspection PyDeprecation
                _import_data_settings_deprecation_warning("registration_image_folder")
            if data_branch.find("registration_image_filename_pattern") is not None:
                # noinspection PyDeprecation
                _import_data_settings_deprecation_warning("registration_image_filename_pattern")
            if data_branch.find("roi_list_path") is not None:
                # noinspection PyDeprecation
                _import_data_settings_deprecation_warning("roi_list_path")
            if data_branch.find("divide_disconnected_roi") is not None:
                # noinspection PyDeprecation
                _import_data_settings_deprecation_warning("divide_disconnected_roi")
            if data_branch.find("extraction_config") is not None:
                # noinspection PyDeprecation
                _import_data_settings_deprecation_warning("extraction_config")

            data_arguments_list += [dict(current_data_arguments)]

    return data_arguments_list


def _import_data_settings_deprecation_warning(tag: str) -> None:
    warnings.warn(
        f"The {tag} tag was deprecated in version 2.0, and will not be parsed.",
        DeprecationWarning
    )



def get_roi_labels(**kwargs):
    raise RuntimeError(
        f"The get_roi_labels function has been replaced by "
        f"mirp.extract_mask_labels."
    )


def get_image_acquisition_parameters(**kwargs):
    raise RuntimeError(
        f"The get_image_acquisition_parameters function has been replaced by "
        f"mirp.extract_image_parameters."
    )


def get_file_structure_parameters(**kwargs):
    raise RuntimeError(
        f"The get_file_structure_parameters function has been fully deprecated, without replacement."
    )


def parse_file_structure(**kwargs):
    raise RuntimeError(
        f"The parse_file_structure function has been fully deprecated, without replacement."
    )


def extract_images_for_deep_learning(**kwargs):
    raise RuntimeError(
        f"The extract_images_for_deep_learning function has been replaced by "
        f"mirp.deep_learning_preprocessing."
    )


def extract_features(**kwargs):
    raise RuntimeError(
        f"The extract_features function has been replaced by mirp.extract_features."
    )


def extract_images_to_nifti(**kwargs):
    raise RuntimeError(
        f"The extract_images_to_nifti function has been replaced by mirp.extract_images."
    )


def process_images(**kwargs):
    raise RuntimeError(
        "The process_images function has been replaced by mirp.extract_features_and_images."
    )



from typing import Any, Generator
from functools import lru_cache

from mirp.settings.feature_parameters import FeatureExtractionSettingsClass
from mirp._features.base_feature import Feature
from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask


def get_discretisation_parameters(settings: FeatureExtractionSettingsClass) -> Generator[dict[str, Any], None, None]:
    for discretisation_method in settings.discretisation_method:
        if discretisation_method in ["fixed_bin_size", "fixed_bin_size_pyradiomics"]:
            bin_width = settings.discretisation_bin_width
            for current_bin_width in bin_width:
                yield {
                    "discretisation_method": discretisation_method,
                    "bin_width": current_bin_width
                }

        elif discretisation_method in ["fixed_bin_number"]:
            bin_number = settings.discretisation_n_bins
            for current_bin_number in bin_number:
                yield {
                    "discretisation_method": discretisation_method,
                    "bin_number": current_bin_number
                }
        else:
            yield {"discretisation_method": discretisation_method}


class HistogramDerivedFeature(Feature):

    def __init__(
            self,
            discretisation_method: str,
            bin_width: None | float = None,
            bin_number: None | int = None,
            cropping_distance: None | float = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.discretisation_method = discretisation_method
        self.bin_width = bin_width
        self.bin_number = bin_number
        self.cropping_distance = cropping_distance

    @staticmethod
    @lru_cache(maxsize=1)
    def discretise_image(
            image: GenericImage,
            mask: BaseMask,
            discretisation_method: str,
            bin_width: float | None,
            bin_number: int | None,
            cropping_distance: float | None
    ) -> tuple[GenericImage, BaseMask]:
        from mirp._image_processing.cropping import crop
        from mirp._image_processing.discretise_image import discretise_image

        if cropping_distance is not None:
            image, mask = crop(
                image=image,
                masks=mask,
                boundary=cropping_distance,
                in_place=False
            )

        image, mask = discretise_image(
            image=image,
            mask=mask,
            discretisation_method=discretisation_method,
            bin_number=bin_number,
            bin_width=bin_width,
            in_place=False
        )

        return image, mask

    def clear_cache(self):
        super().clear_cache()
        self.discretise_image.cache_clear()

    def update_ibsi_compliance(self):
        if self.discretisation_method == "fixed_bin_size_pyradiomics":
            # pyradiomics with fixed bin size is explicitly not IBSI-compliant.
            self.ibsi_compliant = False

    def _get_discretisation_table_name_element(self) -> list[str]:
        table_elements = []
        if self.discretisation_method is not None:
            if self.discretisation_method == "none":
                pass
            elif self.discretisation_method == "fixed_bin_number":
                table_elements += ["fbn"]
                table_elements += ["n" + str(self.bin_number)]
            elif self.discretisation_method == "fixed_bin_size":
                table_elements += ["fbs"]
                table_elements += ["w" + str(self.bin_width)]
            elif self.discretisation_method == "fixed_bin_size_pyradiomics":
                table_elements += ["fbsp"]
                table_elements += ["w" + str(self.bin_width)]
            else:
                raise ValueError(f"Unknown discretisation method: {self.discretisation_method}")

        return table_elements



from mirp._features.histogram import HistogramDerivedFeature


class FeatureTexture(HistogramDerivedFeature):

    def __init__(self, spatial_method: str, **kwargs):
        super().__init__(**kwargs)
        self.spatial_method = spatial_method.lower()

    def _get_spatial_table_name_element(self) -> list[str | None]:
        if self.spatial_method == "2d_average":
            table_elements = ["2d_avg"]
        elif self.spatial_method == "2d_slice_merge":
            table_elements = ["2d_s_mrg"]
        elif self.spatial_method == "2.5d_direction_merge":
            table_elements = ["2.5d_d_mrg"]
        elif self.spatial_method == "2.5d_volume_merge":
            table_elements = ["2.5d_v_mrg"]
        elif self.spatial_method == "3d_average":
            table_elements = ["3d_avg"]
        elif self.spatial_method == "3d_volume_merge":
            table_elements = ["3d_v_mrg"]
        elif self.spatial_method is None:
            table_elements = []
        else:
            table_elements = [self.spatial_method]

        return table_elements



import hashlib

import numpy as np
import pandas as pd
import copy

from mirp._images.generic_image import GenericImage
from mirp._masks.base_mask import BaseMask
from mirp.utilities.utilities import zero_if_close


class Data3DMesh(object):

    def __init__(self):

        # Mesh faces
        self.mesh_faces: np.ndarray | None = None

        # Mesh vertices
        self.mesh_vertices: np.ndarray | None = None

        # Volume
        self.volume: float | None = None

        # Area
        self.area: float | None = None

        # Voxels within intensity mask and morphological mask
        self.data_int: pd.DataFrame | None = None
        self.data_morph: pd.DataFrame | None = None

        # Image spacing
        self.spacing: tuple[float, float, float] | None = None

    def compute(self, image: GenericImage, mask: BaseMask):
        # Skip processing if input image and/or roi are missing
        if image is None or mask is None:
            return

        # Check if data actually exists
        if image.is_empty() or mask.roi_morphology.is_empty_mask():
            return

        from skimage.measure import marching_cubes, mesh_surface_area

        # Image spacing
        self.spacing = image.image_spacing

        # Define tables based on morphological and intensity masks
        data = mask.as_pandas_dataframe(
            image=image,
            intensity_mask=True,
            morphology_mask=True
        )

        self.data_int = data[data.roi_int_mask].reset_index()
        self.data_morph = data[data.roi_morph_mask].reset_index()

        # Get ROI and pad with empty voxels
        morphology_mask = np.pad(
            mask.roi_morphology.get_voxel_grid(),
            pad_width=1,
            mode="constant",
            constant_values=0.0
        )

        # Use marching cubes to generate a mesh grid for the ROI
        vertices, faces, _, _ = marching_cubes(
            volume=morphology_mask,
            level=0.5,
            spacing=mask.roi_morphology.image_spacing
        )

        self.mesh_vertices = vertices
        self.mesh_faces = faces

        # Get vertices for each face
        vert_a = vertices[faces[:, 0], :]
        vert_b = vertices[faces[:, 1], :]
        vert_c = vertices[faces[:, 2], :]

        # noinspection PyUnreachableCode
        self.volume = np.abs(np.sum(
            1.0 / 6.0 * np.einsum("ij,ij->i", vert_a, np.cross(vert_b, vert_c, 1, 1))
        ))

        # noinspection PyUnreachableCode
        self.area = mesh_surface_area(verts=vertices, faces=faces)

    def is_empty(self):
        return self.volume is None

    def is_singular(self):
        if self.is_empty():
            return True

        return len(self.data_morph) <= 1


class Data3DConvexHull(Data3DMesh):

    def __init__(self):
        super().__init__()

        self.convex_hull_vertices: np.ndarray | None = None
        self.convex_hull_volume: float | None = None
        self.convex_hull_area: float | None = None

    def compute_convex_hull(self):
        if self.is_empty():
            return

        from scipy.spatial import ConvexHull

        # Generate the convex hull from the mesh vertices
        convex_hull = ConvexHull(self.mesh_vertices)

        # Extract convex hull vertices, area and volume.
        self.convex_hull_vertices = self.mesh_vertices[convex_hull.vertices, :] - np.mean(self.mesh_vertices, axis=0)
        self.convex_hull_area = convex_hull.area
        self.convex_hull_volume = convex_hull.volume


class Data3DAxisAlignedBoundingBox(Data3DConvexHull):

    def __init__(self):
        super().__init__()

        self.bounding_box_volume: float | None = None
        self.bounding_box_area: float | None = None

    def compute_bounding_box(self):
        if self.is_empty():
            return

        dims = np.max(self.convex_hull_vertices, axis=0) - np.min(self.convex_hull_vertices, axis=0)
        self.bounding_box_volume = np.prod(dims)
        self.bounding_box_area = 2.0 * dims[0] * dims[1] + 2.0 * dims[0] * dims[2] + 2.0 * dims[1] * dims[2]


class Data3DOrientedMinimumBoundingBox(Data3DConvexHull):

    def __init__(self):
        super().__init__()

        self.bounding_box_volume: float | None = None
        self.bounding_box_area: float | None = None

    def compute_bounding_box(self):
        if self.is_empty():
            return

        rot_df = pd.DataFrame({
            "rot_axis_0": np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2]),
            "rot_axis_1": np.array([1, 2, 1, 2, 0, 2, 0, 2, 0, 1, 0, 1]),
            "rot_axis_2": np.array([2, 1, 0, 0, 2, 0, 1, 1, 1, 0, 2, 2]),
            "aabb_axis_0": np.zeros(12),
            "aabb_axis_1": np.zeros(12),
            "aabb_axis_2": np.zeros(12),
            "vol": np.zeros(12)
        })

        # Rotate over different sequences
        for ii in np.arange(0, len(rot_df)):
            # Create a local copy
            work_pos = copy.deepcopy(self.convex_hull_vertices)

            # Rotate over sequence of rotation axes
            work_pos = self._get_optimally_rotated_volume(input_pos=work_pos, rot_axis=rot_df.rot_axis_0[ii])
            work_pos = self._get_optimally_rotated_volume(input_pos=work_pos, rot_axis=rot_df.rot_axis_1[ii])
            work_pos = self._get_optimally_rotated_volume(input_pos=work_pos, rot_axis=rot_df.rot_axis_2[ii])

            # Determine resultant minimum bounding box
            aabb_dims = np.max(work_pos, axis=0) - np.min(work_pos, axis=0)
            rot_df.loc[ii, "aabb_axis_0"] = aabb_dims[0]
            rot_df.loc[ii, "aabb_axis_1"] = aabb_dims[1]
            rot_df.loc[ii, "aabb_axis_2"] = aabb_dims[2]
            rot_df.loc[ii, "vol"] = np.prod(aabb_dims)

            del work_pos, aabb_dims

        # Find minimal volume of all rotations and return bounding box dimensions
        sel_row = rot_df.loc[rot_df.vol.idxmin(), :]
        dims = np.array([sel_row.aabb_axis_0, sel_row.aabb_axis_1, sel_row.aabb_axis_2])

        self.bounding_box_volume = np.prod(dims)
        self.bounding_box_area = 2.0 * dims[0] * dims[1] + 2.0 * dims[0] * dims[2] + 2.0 * dims[1] * dims[2]

    def _get_optimally_rotated_volume(
            self,
            input_pos: np.ndarray,
            rot_axis: int,
            n_minima: int = 3,
            res_init: float = 5.0
    ) -> np.ndarray:
        from scipy.spatial import ConvexHull

        # Find axis aligned bounding box of the point set
        aabb_max = np.max(input_pos, axis=0)
        aabb_min = np.min(input_pos, axis=0)

        # Center the point set at the AABB center
        output_pos = input_pos - 0.5 * (aabb_min + aabb_max)

        # Project model to plane
        proj_pos = copy.deepcopy(output_pos)
        proj_pos = np.delete(proj_pos, [rot_axis], axis=1)

        # Calculate 2D convex hull of the model projection in plane
        if np.shape(proj_pos)[0] >= 10:
            hull_2d = ConvexHull(points=proj_pos)
            hull_mat = proj_pos[hull_2d.vertices, :]
            del hull_2d, proj_pos
        else:
            hull_mat = proj_pos
            del proj_pos

        # Transpose hull_mat so that the array is (ndim, npoints) instead of (npoints, ndim)
        hull_mat = np.transpose(hull_mat)

        # Calculate bounding box surface of a series of rotated contours
        # Note we can program a min-search algorithm here as well

        # Calculate initial surfaces
        theta_init = np.arange(start=0.0, stop=90.0 + res_init, step=res_init) * np.pi / 180.0
        rot_area = np.array(list(map(
            lambda x: self._get_rotated_axis_aligned_bounding_box_area(theta=x, hull_mat=hull_mat), theta_init
        )))

        # Find local minima
        df_min = self._find_peaks(x=rot_area, direction="neg")

        # Check if any minimum was generated
        if len(df_min) > 0:
            # Investigate up to n_minima number of local minima, starting with the global minimum
            df_min = df_min.sort_values(by="val", ascending=True)

            # Determine max number of minima evaluated
            max_iter = np.min([n_minima, len(df_min)])

            # Initialise placeholder array
            theta_min = np.zeros(max_iter)

            # Iterate over local minima
            for k in np.arange(0, max_iter):
                # Find initial angle corresponding to i-th minimum
                sel_ind = df_min.ind.values[k]
                theta_curr = theta_init[sel_ind]

                # Zoom in to improve the approximation of theta
                theta_min[k] = self._find_optimal_rotation_angle(
                    hull_mat=hull_mat,
                    theta_sel=theta_curr,
                    res=res_init * np.pi / 180.0
                )

            # Calculate surface areas corresponding to theta_min and theta that minimises the surface
            rot_area = np.array(list(map(
                lambda x: self._get_rotated_axis_aligned_bounding_box_area(theta=x, hull_mat=hull_mat), theta_min
            )))
            theta_sel = theta_min[np.argmin(rot_area)]

        else:
            theta_sel = theta_init[0]

        # Rotate original point along the angle that minimises the projected AABB area
        output_pos = np.transpose(output_pos)
        rot_mat = self._get_rotation_matrix(theta=theta_sel, dim=3, rot_axis=rot_axis)
        output_pos = np.dot(rot_mat, output_pos)

        # Rotate output_pos back to (npoints, ndim)
        output_pos = np.transpose(output_pos)

        return output_pos

    def _get_rotated_axis_aligned_bounding_box_area(self, theta, hull_mat):
        # Function to calculate surface of the axis-aligned bounding box of a rotated 2D contour

        # Create rotation matrix and rotate over theta
        rot_mat = self._get_rotation_matrix(theta=theta, dim=2)
        rot_hull = np.dot(rot_mat, hull_mat)

        # Calculate bounding box surface of the rotated contour
        rot_aabb_dims = np.max(rot_hull, axis=1) - np.min(rot_hull, axis=1)
        rot_aabb_area = np.prod(rot_aabb_dims)

        return rot_aabb_area

    def _find_optimal_rotation_angle(self, hull_mat, theta_sel, res, max_rep=5):
        # Iterative approximator for finding angle theta that minimises surface area
        for jj in np.arange(0, max_rep):

            # Select new thetas in vicinity of
            theta = np.array([theta_sel-res, theta_sel-0.5*res, theta_sel, theta_sel+0.5*res, theta_sel+res])

            # Calculate projection areas for current angles theta
            rot_area = np.array(list(map(
                lambda x: self._get_rotated_axis_aligned_bounding_box_area(theta=x, hull_mat=hull_mat), theta)
            ))

            # Find global minimum and corresponding angle theta_sel
            theta_sel = theta[np.argmin(rot_area)]

            # Shrink resolution and iterate
            res /= 2.0

        return theta_sel

    @staticmethod
    def _get_rotation_matrix(theta, dim=2, rot_axis=-1):
        # Creates a 2d or 3d rotation matrix
        if dim == 2:
            rot_mat = np.array([
                [np.cos(theta), -np.sin(theta)],
                [np.sin(theta), np.cos(theta)]
            ])

        elif dim == 3:
            if rot_axis == 0:
                rot_mat = np.array([
                    [1.0, 0.0, 0.0],
                    [0.0, np.cos(theta), -np.sin(theta)],
                    [0.0, np.sin(theta), np.cos(theta)]
                ])

            elif rot_axis == 1:
                rot_mat = np.array([
                    [np.cos(theta), 0.0, np.sin(theta)],
                    [0.0, 1.0, 0.0],
                    [-np.sin(theta), 0.0, np.cos(theta)]
                ])

            elif rot_axis == 2:
                rot_mat = np.array([
                    [np.cos(theta), -np.sin(theta), 0.0],
                    [np.sin(theta), np.cos(theta), 0.0],
                    [0.0, 0.0, 1.0]
                ])
            else:
                rot_mat = None
        else:
            rot_mat = None

        return rot_mat

    def _find_peaks(self, x, direction="pos"):
        # Determines peak positions in array of values

        # Invert when looking for local minima
        if direction == "neg":
            x = -x

        # Generate segments where slope is negative
        df_segm = self._segmentise_input(x=(np.diff(x) < 0.0) * 1)

        # Start of slope coincides with position of peak (due to index shift induced by np.diff)
        ind_peak = df_segm.loc[df_segm.val == 1, "i"].values

        # Check right boundary
        if x[-1] > x[-2]:
            ind_peak = np.append(ind_peak, np.shape(x)[0]-1)

        # Construct dataframe with index and corresponding value
        if np.shape(ind_peak)[0] == 0:
            df_peak = pd.DataFrame({"ind": [], "val": []})
        else:
            if direction == "pos":
                # Positive direction
                df_peak = pd.DataFrame({"ind": ind_peak, "val": x[ind_peak]})
            else:
                # Negative direction
                df_peak = pd.DataFrame({"ind": ind_peak, "val": -x[ind_peak]})

        return df_peak

    @staticmethod
    def _segmentise_input(x):
        # Produces a list of segments from input x with values (0,1)

        # Create a difference vector
        y = np.diff(x)

        # Find start and end indices of sections with value 1
        ind_1_start = np.array(np.where(y == 1)).flatten()
        if np.shape(ind_1_start)[0] > 0:
            ind_1_start += 1
        ind_1_end = np.array(np.where(y == -1)).flatten()

        # Check for boundary effects
        if x[0] == 1:
            ind_1_start = np.insert(ind_1_start, 0, 0)
        if x[-1] == 1:
            ind_1_end = np.append(ind_1_end, np.shape(x)[0] - 1)

        # Generate segment df for segments with value 1
        if np.shape(ind_1_start)[0] == 0:
            df_one = pd.DataFrame({
                "i": [],
                "j": [],
                "val": []
            })
        else:
            df_one = pd.DataFrame({
                "i": ind_1_start,
                "j": ind_1_end,
                "val": np.ones(np.shape(ind_1_start)[0])
            })

        # Find start and end indices for section with value 0
        if np.shape(ind_1_start)[0] == 0:
            ind_0_start = np.array([0])
            ind_0_end = np.array([np.shape(x)[0] - 1])

        else:
            ind_0_end = ind_1_start - 1
            ind_0_start = ind_1_end + 1

            # Check for boundary effect
            if x[0] == 0:
                ind_0_start = np.insert(ind_0_start, 0, 0)
            if x[-1] == 0:
                ind_0_end = np.append(ind_0_end, np.shape(x)[0] - 1)

            # Check for out-of-range boundary effects
            if ind_0_end[0] < 0:
                ind_0_end = np.delete(ind_0_end, 0)
            if ind_0_start[-1] >= np.shape(x)[0]:
                ind_0_start = np.delete(ind_0_start, -1)

        # Generate segment df for segments with value 0
        if np.shape(ind_0_start)[0] == 0:
            df_zero = pd.DataFrame({
                "i": [],
                "j": [],
                "val": []
            })

        else:
            df_zero = pd.DataFrame({
                "i": ind_0_start,
                "j": ind_0_end,
                "val": np.zeros(np.shape(ind_0_start)[0])
            })

        df_segm = pd.concat(
            [df_one, df_zero],
            ignore_index=True
        ).sort_values(by="i").reset_index(drop=True)

        return df_segm


class Data3DPrincipleComponents(Data3DMesh):

    def __init__(self):
        super().__init__()

        self.semi_axes: tuple[float, float, float] | None = None

    def compute_semi_axes(self):
        if self.is_empty():
            return

        # Get position matrix
        pos_mat_pca = self.data_morph[["z", "y", "x"]].values

        # Subtract mean
        pos_mat_pca = np.multiply(
            (pos_mat_pca - np.mean(pos_mat_pca, axis=0)),
            self.spacing
        )

        # Get eigenvalues and vectors
        if not self.is_singular():
            eigen_val, eigen_vec = np.linalg.eigh(np.cov(pos_mat_pca, rowvar=False))
            eigen_val = zero_if_close(eigen_val)
            self.semi_axes = tuple(2.0 * np.sqrt(np.sort(eigen_val)))

    def get_ellipsoid_surface_area(self, n_degree=10):
        # Approximates area of an ellipsoid using legendre polynomials

        # Let semi_axes[2] be the major semi-axis length, semi_axes[1] the minor semi-axis length and semi_axes[0] the
        # least axis length.

        # Import legendre evaluation function from numpy
        from numpy.polynomial.legendre import legval

        # Check if the semi-axes differ in length, otherwise the ellipsoid is spherical and legendre polynomials are not
        # required.
        if self.semi_axes[0] == self.semi_axes[1] and self.semi_axes[0] == self.semi_axes[2]:
            # Exact sphere calculation
            area_appr = 4.0 * np.pi * self.semi_axes[0] ** 2.0

        elif self.semi_axes[0] == self.semi_axes[1]:
            # Exact prolate spheroid (major semi-axis > other, equally long, semi-axes)
            ecc = np.sqrt(1.0 - (self.semi_axes[0] ** 2.0) / (self.semi_axes[2] ** 2.0))

            # Calculate area
            area_appr = (
                    2.0 * np.pi * self.semi_axes[0] ** 2.0
                    * (1.0 + self.semi_axes[2] * np.arcsin(ecc) / (self.semi_axes[0] * ecc))
            )

        elif self.semi_axes[1] == self.semi_axes[2]:
            # Exact oblate spheroid (major semi-axes equally long > shortest semi-axis)
            ecc = np.sqrt(1.0 - (self.semi_axes[0] ** 2.0) / (self.semi_axes[2] ** 2.0))

            # Calculate area:
            area_appr = 2.0 * np.pi * self.semi_axes[2] ** 2.0 * (1.0 + ((1.0 - ecc ** 2.0) / ecc) * np.arctanh(ecc))

        else:
            # Tri-axial ellipsoid
            # Calculate eccentricities
            ecc_alpha = np.sqrt(1.0 - (self.semi_axes[1] ** 2.0) / (self.semi_axes[2] ** 2.0))
            ecc_beta = np.sqrt(1.0 - (self.semi_axes[0] ** 2.0) / (self.semi_axes[2] ** 2.0))

            # Create a vector to generate coefficients for the Legendre polynomial
            nu = np.arange(start=0, stop=n_degree + 1, step=1) * 1.0

            # Define Legendre polynomial coefficients and evaluation point
            leg_coeff = (ecc_alpha * ecc_beta) ** nu / (1.0 - 4.0 * nu ** 2.0)
            leg_x = (ecc_alpha ** 2.0 + ecc_beta ** 2.0) / (2.0 * ecc_alpha * ecc_beta)

            # Calculate approximate area
            area_appr = 4.0 * np.pi * self.semi_axes[2] * self.semi_axes[1] * legval(x=leg_x, c=leg_coeff)

        return area_appr


class Data3DMinimumEnvelopingEllipsoid(Data3DConvexHull, Data3DPrincipleComponents):

    def __init__(self):
        super().__init__()

    def compute_semi_axes(self, tolerance: float = 10E-4):
        if self.is_empty() or self.is_singular():
            return

        import numpy.linalg as npla

        # Cast to np.matrix
        mat_p = copy.deepcopy(self.convex_hull_vertices)
        mat_p = np.transpose(mat_p)

        # Dimension of the point set
        ndim = np.shape(mat_p)[0]

        # Number of points in the point set
        npoint = np.shape(mat_p)[1]

        # Add a row of 1s to the ndim x npoint matrix input_pos - so mat_q is (ndim+1) x npoint now.
        mat_q = np.append(mat_p, np.ones(shape=(1, npoint)), axis=0)

        # Initialise settings for Khachiyan algorithm
        iter_count = 1
        err = 1.0

        # Initial u vector
        vec_u = 1.0 * np.ones(shape=npoint) / npoint

        # Khachiyan algorithm
        while err > tolerance:
            # Matrix multiplication:
            # np.diag(u) : if u is a vector, places the elements of u
            # in the diagonal of an NxN matrix of zeros
            mat_x = np.matmul(mat_q, np.matmul(np.diag(vec_u), np.transpose(mat_q)))

            # npla.inv(mat_x) returns the matrix inverse of mat_x
            # np.diag(mat_M) when mat_M is a matrix returns the diagonal vector of mat_M
            vec_m = np.diag(np.matmul(np.transpose(mat_q), np.matmul(npla.inv(mat_x), mat_q)))

            # Find the value and location of the maximum element in the vector M
            max_m = np.max(vec_m)
            ind_j = np.argmax(vec_m)

            # Calculate the step size for the ascent
            step_size = (max_m - ndim - 1.0) / ((ndim + 1.0) * (max_m - 1.0))

            # Calculate the vector vec_new_u:
            # Take the vector vec_u, and multiply all the elements in it by (1-step_size)
            vec_new_u = vec_u * (1.0 - step_size)

            # Increment the jth element of new_u by step_size
            vec_new_u[ind_j] += step_size

            # Store the error by taking finding the square root of the SSD between new_u and u. The SSD or sum-of-square
            # differences, takes two vectors of the same size, creates a new vector by finding the difference between
            # corresponding elements, squaring each difference and adding them all together.
            err = npla.norm(vec_new_u - vec_u)

            # Increment iter_count and replace vec_u
            iter_count += 1
            vec_u = vec_new_u

        # Put the elements of the vector u into the diagonal of a matrix
        # U with the rest of the elements as 0
        mat_u = np.diag(vec_u)

        # Compute the A-matrix
        mat_a = (1.0 / ndim) * npla.inv(
            np.matmul(mat_p, np.matmul(mat_u, np.transpose(mat_p)))
            - np.matmul(np.matmul(mat_p, vec_u), np.transpose(np.matmul(mat_p, vec_u)))
        )

        # Perform singular value decomposition
        s = npla.svd(mat_a, compute_uv=False)

        # The semi-axis lengths are the inverse square root of the singular values
        self.semi_axes = tuple(np.sort(1.0 / np.sqrt(s)))


class Data3DSpatial(Data3DMesh):

    def __init__(self):
        super().__init__()

        self.moran_i: float | None = None
        self.geary_c: float | None = None

    def compute_spatial_information(
            self,
            image: GenericImage,
            mask: BaseMask,
            allow_approximation: bool = True
    ):
        n_v_int = len(self.data_int)

        if (1 < n_v_int < 1000) or not allow_approximation:
            # Calculate geospatial features using a brute force approach
            self.moran_i, self.geary_c = self._spatial_distribution(data=self.data_int)

        elif n_v_int >= 1000:
            # Use monte carlo approach to estimate geospatial features

            m = hashlib.sha1(usedforsecurity=False)
            m = image.update_hash(m=m)
            m = mask.roi_morphology.update_hash(m=m)
            randomiser = np.random.default_rng(int(m.hexdigest(), 16))

            # Create lists for storing feature values
            moran_list, geary_list = [], []

            # Initiate iterations
            iter_nr = 1
            tol_aim = 0.002
            tol_sem = 1.000

            # Iterate until the sample error of the mean drops below the target tol_aim
            while tol_sem > tol_aim:

                # Select a small random subset of 100 points in the volume
                curr_points = randomiser.choice(n_v_int, size=100, replace=False)

                # Calculate Moran's I and Geary's C for the point subset
                moran_i, geary_c = self._spatial_distribution(data=self.data_int.loc[curr_points, :])

                # Append values to the lists
                moran_list.append(moran_i)
                geary_list.append(geary_c)

                # From the tenth iteration, estimate the sample error of the mean
                if iter_nr > 10:
                    tol_sem = np.max([np.std(moran_list), np.std(geary_list)]) / np.sqrt(iter_nr)

                # Update counter
                iter_nr += 1

            self.moran_i = np.mean(moran_list)
            self.geary_c = np.mean(geary_list)

    def _spatial_distribution(self, data: pd.DataFrame):
        n_v = len(data)

        # Generate point cloud
        pos_mat = data[["z", "y", "x"]].values
        pos_mat = np.multiply(pos_mat, self.spacing)

        if n_v < 2000:
            # Determine all interactions between voxels
            comb_iter = np.array([np.tile(np.arange(0, n_v), n_v), np.repeat(np.arange(0, n_v), n_v)])
            comb_iter = comb_iter[:, comb_iter[0, :] > comb_iter[1, :]]

            # Determine weighting for all interactions (inverse weighting with distance)
            w_ij = 1.0 / np.array(list(
                map(lambda i: np.sqrt(np.sum((pos_mat[comb_iter[0, i], :] - pos_mat[comb_iter[1, i], :]) ** 2.0)),
                    np.arange(np.shape(comb_iter)[1]))))

            # Create array of mean-corrected grey level intensities
            gl_dev = data.g.values - np.mean(data.g)

            # Moran's I
            nom = n_v * np.sum(np.multiply(np.multiply(w_ij, gl_dev[comb_iter[0, :]]), gl_dev[comb_iter[1, :]]))
            denom = np.sum(w_ij) * np.sum(gl_dev ** 2.0)
            if denom > 0.0:
                moran_i = nom / denom
            else:
                # If the denominator is 0.0, this basically means only one intensity is present in the volume, which
                # indicates ideal spatial correlation.
                moran_i = 1.0

            # Geary's C
            nom = (n_v - 1.0) * np.sum(np.multiply(w_ij, (gl_dev[comb_iter[0, :]] - gl_dev[comb_iter[1, :]]) ** 2.0))
            if denom > 0.0:
                geary_c = nom / (2.0 * denom)
            else:
                # If the denominator is 0.0, this basically means only one intensity is present in the volume.
                geary_c = 1.0
        else:
            # In practice, this code variant is only used if the ROI is too large to perform all distance calculations
            # in one go.

            # Create array of mean-corrected grey level intensities
            gl_dev = data.g.values - np.mean(data.g)

            moran_nom = 0.0
            geary_nom = 0.0
            w_denom = 0.0

            # Iterate over voxels
            for ii in np.arange(n_v - 1):
                # Get all jj > ii voxels
                jj = np.arange(start=ii + 1, stop=n_v)

                # Get distance weights
                w_iijj = 1.0 / np.sqrt(np.sum(np.power(pos_mat[ii, :] - pos_mat[jj, :], 2.0), axis=1))

                moran_nom += np.sum(np.multiply(np.multiply(w_iijj, gl_dev[ii]), gl_dev[jj]))
                geary_nom += np.sum(np.multiply(w_iijj, (gl_dev[ii] - gl_dev[jj]) ** 2.0))
                w_denom += np.sum(w_iijj)

            gl_denom = np.sum(gl_dev ** 2.0)

            # Moran's I index
            if gl_denom > 0.0:
                moran_i = n_v * moran_nom / (w_denom * gl_denom)
            else:
                moran_i = 1.0

            # Geary's C measure
            if gl_denom > 0.0:
                geary_c = (n_v - 1.0) * geary_nom / (2 * w_denom * gl_denom)
            else:
                geary_c = 1.0

        return moran_i, geary_c



import os
import os.path
from functools import singledispatch
import numpy as np
import pandas as pd

from mirp._data_import.directory import ImageDirectory, MaskDirectory
from mirp._data_import.generic_file import ImageFile, MaskFile
from mirp._data_import.utilities import supported_file_types, supported_image_modalities, flatten_list
from mirp._images.base_image import BaseImage
from mirp._masks.base_mask import BaseMask
from mirp.settings.import_data_parameters import import_data_settings


def import_image(
        image,
        sample_name: None | str | list[str] = None,
        image_name: None | str | list[str] = None,
        image_file_type: None | str = None,
        image_modality: None | str | list[str] = None,
        image_sub_folder: None | str = None,
        stack_images: str = "auto"
) -> list[ImageFile]:
    """
    Creates and curates references to image files.

    Parameters
    ----------
    image: Any
        A path to an image file, a path to a directory containing image files, a path to a config_data.xml
        file, a path to a csv file containing references to image files, a pandas.DataFrame containing references to
        image files, or a numpy.ndarray.

    sample_name: str or list of str, default: None
        Name of expected sample names. This is used to select specific image files. If None, no image files are
        filtered based on the corresponding sample name (if known).

    image_name: str, optional, default: None
        Pattern to match image files against. The matches are exact. Use wildcard symbols ("*") to
        match varying structures. The sample name (if part of the file name) can also be specified using "#". For
        example, image_name = '#_*_image' would find John_Doe in John_Doe_CT_image.nii or John_Doe_001_image.nii.
        File extensions do not need to be specified. If None, file names are not used for filtering files and
        setting sample names.

    image_file_type: {"dicom", "nifti", "nrrd", "numpy", "itk"}, optional, default: None
        The type of file that is expected. If None, the file type is not used for filtering files.
        "itk" comprises "nifti" and "nrrd" file types.

    image_modality: {"ct", "pet", "pt", "mri", "mr", "rtdose", "generic"}, optional, default: None
        The type of modality that is expected. If None, modality is not used for filtering files. Note that only
        DICOM files contain metadata concerning modality.

    image_sub_folder: str, optional, default: None
        Fixed directory substructure where image files are located. If None, the directory substructure is not used
        for filtering files.

    stack_images: {"auto", "yes", "no"}, optional, default: "str"
        If image files in the same directory cannot be assigned to different samples, and are 2D (slices) of the same
        size, they might belong to the same 3D image stack. "auto" will stack 2D numpy arrays, but not other file types.
        "yes" will stack all files that contain 2D images, that have the same dimensions, orientation and spacing,
        except for DICOM files. "no" will not stack any files. DICOM files ignore this argument, because their stacking
        can be determined from metadata.

    Returns
    -------
    list of ImageFile
        The functions returns a list of ImageFile objects, if any were found with the specified filters.
    """
    # Check modality.
    if image_modality is not None:
        if not isinstance(image_modality, str):
            raise TypeError(
                f"The image_modality argument is expected to be a single character string or None. The following "
                f"modalities are supported: {', '.join(supported_image_modalities(None))}.")
        _ = supported_image_modalities(image_modality.lower())

    # Check image_file_type.
    if image_file_type is not None:
        if not isinstance(image_file_type, str):
            raise TypeError(
                f"The image_file_type argument is expected to be a single character string, or None. The following "
                f" file types are supported: {', '.join(supported_file_types(None))}.")
        _ = supported_file_types(image_file_type)

    # Check stack_images
    if stack_images not in ["yes", "auto", "no"]:
        raise ValueError(
            f"The stack_images argument is expected to be one of yes, auto, or no. Found: {stack_images}."
        )

    image_list: ImageFile | list[ImageFile] = _import_image(
        image,
        sample_name=sample_name,
        image_name=image_name,
        image_file_type=image_file_type,
        image_modality=image_modality,
        image_sub_folder=image_sub_folder,
        stack_images=stack_images,
        is_mask=False
    )

    if not isinstance(image_list, list):
        image_list = [image_list]

    # Flatten list.
    image_list = flatten_list(image_list)

    replacement_image_modality = None
    if isinstance(image_modality, str):
        replacement_image_modality = image_modality
    elif isinstance(image_modality, list) and len(image_modality) == 1:
        replacement_image_modality = image_modality[1]

    for image in image_list:
        image.set_modality(modality=replacement_image_modality)

    return image_list


@singledispatch
def _import_image(image, **kwargs):
    raise NotImplementedError(f"Unsupported image type: {type(image)}")


@_import_image.register(list)
def _(image: list, **kwargs):
    # List can be anything. Hence, we dispatch import_image for the individual list elements.
    image_list = [_import_image(current_image, **kwargs) for current_image in image]

    return image_list


@_import_image.register(str)
def _(
        image: str,
        is_mask: bool = False,
        **kwargs
):
    # Image is a string, which could be a path to a xml file, to a csv file, or just a regular
    # path a path to a file, or a path to a directory. Test which it is and then dispatch.

    if image.lower().endswith("xml"):
        data_arguments = import_data_settings(path=image, is_mask=is_mask)
        image_list = []

        for current_data_arguments in data_arguments:
            image = current_data_arguments["image"]
            current_data_arguments.pop("image")
            image_list += [_import_image(image, is_mask=is_mask, **current_data_arguments)]

        return image_list

    elif image.lower().endswith("csv"):
        ...

    elif os.path.isdir(image):
        if is_mask:
            return _import_image(MaskDirectory(directory=image, **kwargs), remove_metadata=True)
        else:
            return _import_image(ImageDirectory(directory=image, **kwargs), remove_metadata=True)

    elif os.path.exists(image):
        if is_mask:
            return _import_image(MaskFile(file_path=image, **kwargs), remove_metadata=True)
        else:
            return _import_image(ImageFile(file_path=image, **kwargs), remove_metadata=True)

    else:
        raise ValueError("The image path does not point to a xml file, a csv file, a valid image file or a directory "
                         "containing imaging.")


@_import_image.register(pd.DataFrame)
def _(
        image: pd.DataFrame,
        image_modality: None | str = None,
        **kwargs
):
    ...


@_import_image.register(np.ndarray)
def _(
        image: np.ndarray,
        is_mask: bool = False,
        **kwargs
):

    from mirp._data_import.numpy_file import ImageNumpyFile, MaskNumpyFile

    if is_mask:
        image_object = MaskNumpyFile(**kwargs)
    else:
        image_object = ImageNumpyFile(**kwargs)

    image_object.image_data = image_object.image_metadata = image
    image_object.complete()
    image_object.update_image_data()
    image_object.check(raise_error=True, remove_metadata=False)

    return image_object


@_import_image.register(ImageFile)
def _(
        image: ImageFile,
        remove_metadata: bool = False,
        **kwargs
):

    # Create image.
    image = image.create()

    # Check if the data are consistent.
    image.check(raise_error=True)

    # Complete image data and add identifiers (if any)
    image.complete()

    if remove_metadata:
        image.remove_metadata()

    return image


@_import_image.register(ImageDirectory)
def _(
        image: ImageDirectory,
        remove_metadata: bool = False,
        **kwargs
):

    # Check first if the data are consistent for a directory.
    image.check(raise_error=True)

    # Yield image files.
    image.create_images()

    # Dispatch to import_image method for ImageFile objects. This performs a last check and completes the object.
    return [
        _import_image(current_image, remove_metadata=remove_metadata, **kwargs)
        for current_image in image.image_files
    ]


@_import_image.register(BaseImage)
def _(
        image: BaseImage,
        **kwargs
):
    return image


@_import_image.register(BaseMask)
def _(
        image: BaseMask,
        **kwargs
):
    return image


from typing import Any

from mirp._data_import.dicom_multi_frame import ImageDicomMultiFrame
from mirp._data_import.dicom_file_mr import ImageDicomFileMR
from mirp._data_import.utilities import get_pydicom_meta_tag


class ImageDicomFileMRADC(ImageDicomFileMR):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def export_metadata(self, self_only=False, **kwargs) -> None | dict[str, Any]:
        if not self_only:
            metadata = super().export_metadata()
        else:
            metadata = {}

        self.load_metadata()

        dcm_meta_data = []

        # Diffusion b-value
        b_value = get_pydicom_meta_tag(
            dcm_seq=self.image_metadata,
            tag=(0x0018, 0x9087),
            tag_type="float",
            macro_dcm_seq=(0x0018, 0x9117)
        )
        if b_value is not None:
            dcm_meta_data += [("diffusion_b_value", b_value)]

        metadata.update(dict(dcm_meta_data))
        return metadata


class ImageDicomFileMRADCMultiFrame(ImageDicomMultiFrame, ImageDicomFileMRADC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


import datetime
import logging
from copy import deepcopy

import numpy as np

from mirp._data_import.utilities import convert_dicom_time, get_pydicom_meta_tag, set_pydicom_meta_tag


class SUVscalingObj:  # pragma: no cover

    def __init__(self, dcm):

        # Start of image acquisition for the current position
        acquisition_start_date = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0008, 0x0022), tag_type="str")
        acquisition_start_time = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0008, 0x0032), tag_type="str")
        self.acquisition_ref_time = convert_dicom_time(date_str=acquisition_start_date, time_str=acquisition_start_time)

        # Frame reference time frame (ms)
        self.frame_duration = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0018, 0x1242), tag_type="float")
        self.frame_reference_time = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1300), tag_type="float")

        # Radionuclide administration
        if get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x0016), test_tag=True):
            radio_admin_start_time = get_pydicom_meta_tag(dcm_seq=dcm[0x0054, 0x0016][0], tag=(0x0018, 0x1078), tag_type="str")
            self.radio_admin_ref_time = convert_dicom_time(datetime_str=radio_admin_start_time)

            if self.radio_admin_ref_time is None:
                # If unsuccessful, attempt determining administration time from (0x0018, 0x1072)
                radio_admin_start_time = get_pydicom_meta_tag(dcm_seq=dcm[0x0054, 0x0016][0], tag=(0x0018, 0x1072), tag_type="str")
                self.radio_admin_ref_time = convert_dicom_time(date_str=acquisition_start_date, time_str=radio_admin_start_time)
        else:
            self.radio_admin_ref_time = None

        if self.radio_admin_ref_time is None:
            # If neither (0x0018, 0x1078) or (0x0018, 0x1072) are present, attempt to read private tags

            # GE tags
            ge_acquistion_ref_time = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0009, 0x100d), tag_type="str")
            ge_radio_admin_ref_time = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0009, 0x103b), tag_type="str")

            if ge_acquistion_ref_time is not None and ge_radio_admin_ref_time is not None:
                self.acquisition_ref_time = convert_dicom_time(datetime_str=ge_acquistion_ref_time)
                self.radio_admin_ref_time = convert_dicom_time(datetime_str=ge_radio_admin_ref_time)

        if self.radio_admin_ref_time is not None and self.acquisition_ref_time is not None:

            day_diff = abs(self.radio_admin_ref_time - self.acquisition_ref_time).days
            if day_diff > 1:
                # Correct for de-identification mistakes (i.e. administration time was de-identified correctly, but acquisition time not)
                # We do not expect that the difference between the two is more than a day, or even more than a few hours at most.
                if self.radio_admin_ref_time > self.acquisition_ref_time:
                    self.radio_admin_ref_time -= datetime.timedelta(days=day_diff)
                else:
                    self.radio_admin_ref_time += datetime.timedelta(days=day_diff)

            if self.radio_admin_ref_time > self.acquisition_ref_time + datetime.timedelta(hours=6):
                # Correct for overnight
                self.radio_admin_ref_time -= datetime.timedelta(days=1)

        # Radionuclide total dose and radionuclide half-life
        if get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x0016), test_tag=True):
            self.total_dose = get_pydicom_meta_tag(dcm_seq=dcm[0x0054, 0x0016][0], tag=(0x0018, 0x1074), tag_type="float")
            self.half_life = get_pydicom_meta_tag(dcm_seq=dcm[0x0054, 0x0016][0], tag=(0x0018, 0x1075), tag_type="float")
        else:
            self.total_dose = None
            self.half_life = None

        # Type of intensity in a voxel
        self.voxel_unit = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1001), tag_type="str")

        # Type of decay correction that is used
        self.decay_correction = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1102), tag_type="str", default="NONE")

        # Decay factor for the image
        self.decay_factor = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1321), tag_type="float", default=1.0)

        # TODO: Determine if any decay correction took place from (0018,9758), which has either YES or NO, if present. If YES, (0018,9701) should be present as well.
        # TODO: Use Decay Correction DateTime (0018,9701) as alternative for determining the correction time.

        # Type of SUV
        self.suv_type = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1006), tag_type="str", default="BW")

        # Patient data
        self.patient_gender = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0010, 0x0040), tag_type="str")

        self.patient_height = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0010, 0x1020), tag_type="float")
        if self.patient_height is not None:
            if self.patient_height > 3.0:
                # Interpret patient height as cm and convert to meter
                self.patient_height = self.patient_height / 100.0

        self.patient_weight = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0010, 0x1030), tag_type="float")
        if self.patient_weight is None:
            logging.warning("Patient weight was not found in the DICOM header. SUV normalisation cannot take place.")

        # Private scale factors
        self.philips_suv_scale = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x7053, 0x1000), tag_type="float")
        self.philips_count_scale = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x7053, 0x1009), tag_type="float")

    def get_scale_factor(self, suv_normalisation="bw"):

        if self.voxel_unit == "CNTS":
            # Voxels represent counts
            suv_scale_factor = self.get_scale_factor_counts()
        elif self.voxel_unit == "BQML":
            # Voxels represent Bq/milliliter
            suv_scale_factor = self.get_scale_factor_bqml()
        elif self.voxel_unit in ["GML", "CM2ML"]:
            # Voxels represent a SUV scale: grams/milliliter or cm^2 per milliliter
            suv_scale_factor = 1.0
        else:
            if self.voxel_unit is None:
                logging.warning("SUV scale factor could not be found as the voxel unit tag (0x0054, 0x1001) was missing.")
            else:
                logging.warning("SUV scale factor %s was not implemented.", self.voxel_unit)
            suv_scale_factor = None

        # Parse scale factor
        suv_scale_factor = self.apply_normalisation(scale_factor=suv_scale_factor, norm_method=suv_normalisation)

        return suv_scale_factor

    def get_scale_factor_counts(self):
        """
        SUV scaling for voxels that represent counts (CNTS)
        :return: body-weight corrected scale factor
        """

        if self.philips_suv_scale is not None:
            suv_scale_factor = self.philips_suv_scale
        elif self.philips_count_scale is not None:
            # Determine the decayed dose
            decayed_dose = self.get_decayed_dose()

            if decayed_dose is not None and self.patient_weight is not None:
                suv_scale_factor = self.philips_count_scale * self.patient_weight * 1000.0 / decayed_dose
            else:
                suv_scale_factor = None
        else:
            logging.warning("Counts (CNTS) could not be converted to SUV, because either the Philips scale factor (0x7053, 0x1000) or the Philips count scale (0x7053, "
                            "0x1009) was missing.")
            suv_scale_factor = None

        return suv_scale_factor

    def get_scale_factor_bqml(self):
        """
        SUV scaling for voxels that represent activity
        :return: body-weight corrected scale factor
        """

        decayed_dose = self.get_decayed_dose()

        if decayed_dose is not None and self.patient_weight is not None:
            suv_scale_factor = self.patient_weight * 1000.0 / decayed_dose
        else:
            suv_scale_factor = None

        return suv_scale_factor

    # def get_scale_factor_gml(self):
    #     """
    #     SUV scaling for voxels that already represent a SUV value
    #     :return: body-weight corrected scale factor
    #
    #     """
    #
    #     if self.suv_type == "BW":
    #         # This is the default option
    #         suv_scale_factor = 1.0
    #     elif self.suv_type == "BSA":
    #         # Body surface area
    #         suv_scale_factor = None
    #     elif self.suv_type == "LBM":
    #         # Lean body weight using James' method
    #         suv_scale_factor = None
    #     elif self.suv_type == "LBMJAMES128":
    #         # Lean body weight using James' method with a multiplier of 128 for males
    #         suv_scale_factor = None
    #     elif self.suv_type == "LBMJANMA":
    #         # Lean body weight using the Janmahasatian's method
    #         suv_scale_factor = None
    #     elif self.suv_type == "IBW":
    #         # Ideal body weight
    #         suv_scale_factor = None
    #     else:
    #         suv_scale_factor = None
    #
    #     return suv_scale_factor

    def get_decayed_dose(self):
        """
        Determines dose for normalising the SUV counts
        :return: Dose after decay corrections
        """

        # Check whether required settings are provided
        if self.total_dose is None:
            logging.warning("Radionuclide total dose (0x0018, 0x1074) was not specified in the Radiopharmaceutical information sequence (0x0054, 0x0016).")
            return None

        if self.decay_correction in ["NONE", "START"]:
            if self.frame_duration is None:
                logging.warning("Frame duration (0x0018, 0x1242) is not known.")
                return None

            if self.acquisition_ref_time is None:
                logging.warning("Acquisition date (0x0008, 0x0022) and time (0x0008, 0x0032) are not known.")
                return None

            if self.radio_admin_ref_time is None:
                logging.warning("Time of radionucleitide injection is not known.")
                return None

            if self.half_life is None:
                logging.warning("Radionucleitide half-life (0x0018, 0x1075) in the Radiopharmaceutical information sequence (0x0054, 0x0016) is not known.")
                return None

        if self.decay_correction in ["START"]:
            if self.frame_reference_time is None:
                logging.warning("Frame reference time (0x0054, 0x1300) is not known.")

        # Process for different decay corrections
        if self.decay_correction == "NONE":
            # No decay correction; correct for period between administration and acquisition + 1/2 frame duration
            frame_center_time = self.acquisition_ref_time + datetime.timedelta(microseconds=int(np.round(self.frame_duration * 1000.0 / 2.0)))
            decay_factor = np.power(2.0, (frame_center_time - self.radio_admin_ref_time).seconds / self.half_life)
            decayed_dose = self.total_dose / decay_factor

        elif self.decay_correction == "START":
            # Decay correction of pixel values for the period from pixel acquisition up to scan start
            # Additionally correct for decay between administration and acquisition start

            # Back compute start reference time from acquisition date and time.
            decay_constant = np.log(2) / self.half_life

            # Compute decay during frame. Note that frame duration is converted from ms to s.
            decay_during_frame = decay_constant * self.frame_duration / 1000.0

            # Time at which the average count rate is found.
            time_count_average = 1 / decay_constant * np.log(decay_during_frame / (1.0 - np.exp(-decay_during_frame)))

            # Set reference start time (this may coincide with the series time, but series time may be unreliable).
            reference_start_time = self.acquisition_ref_time + datetime.timedelta(seconds=(time_count_average - self.frame_reference_time / 1000.0))

            # Compute decay time.
            if reference_start_time >= self.radio_admin_ref_time:
                decay_time = (reference_start_time - self.radio_admin_ref_time).seconds
            else:
                decay_time = -(self.radio_admin_ref_time - reference_start_time).seconds

            decay_factor = np.power(2.0, decay_time / self.half_life)
            decayed_dose = self.total_dose / decay_factor

        elif self.decay_correction == "ADMIN":
            # Decay correction of pixel values for the period from pixel acquisition up to administration
            # No additional correction required
            decayed_dose = self.total_dose
            decay_factor = 1.0
        else:
            logging.warning(f"Decay correction (0x0054, 0x1102) was not recognized ({self.decay_correction}) and could not be parsed: %s.")
            decayed_dose = None
            decay_factor = 1.0

        # Update decay factor parameter and decay correction
        self.decay_factor = decay_factor * self.decay_factor
        self.decay_correction = "ADMIN"

        return decayed_dose

    def apply_normalisation(self, scale_factor, norm_method):

        norm_method = deepcopy(norm_method)
        # Translate norm_method to a corresponding DICOM SUV type attribute
        if norm_method in ["bw", "body_weight"]:
            norm_method = "BW"
        elif norm_method in ["bsa", "body_surface_area"]:
            norm_method = "BSA"
        elif norm_method == "lbm_james":
            norm_method = "LBM"
        elif norm_method in ["lbm_janmahasatian", "lbm", "lbm_janma", "lean_body_mass"]:
            norm_method = "LBM_JANMA"
        elif norm_method == "ideal_body_weight":
            norm_method = "IBW"
        else:
            raise ValueError(f"{norm_method} is not a valid SUV normalisation method.")

        # Return scale factor if the normalisation method and the DICOM SUV type attribute match
        if norm_method == self.suv_type:
            return scale_factor

        # Convert the scale factor to a stripped version (PET activity/dose present)
        norm_factor = self.get_norm_factor(suv_type=self.suv_type)

        if norm_factor is None:
            return None
        else:
            scale_factor /= norm_factor

        # Calculate the suv scaling factor using the desired method
        norm_factor = self.get_norm_factor(suv_type=norm_method)

        if norm_factor is None:
            return None
        else:
            scale_factor *= norm_factor

        # Update SUV type and voxel unit
        self.suv_type = norm_method

        return scale_factor

    def get_norm_factor(self, suv_type):
        """
        Calculates SUV scaling factor.
        :param suv_type: type of SUV correction or normalisation used
        :return:
        """

        if suv_type == "BW":
            # Body weight-corrected SUV
            norm_factor = self.patient_weight * 1000.0
        elif suv_type == "BSA":
            # Kim et al. Journal of Nuclear Medicine. Volume 35, No. 1, January 1994. pp 164-167
            norm_factor = self.patient_weight ** 0.425 * (self.patient_height * 100.0) ** 0.725 * 0.007184
        elif suv_type in ["LBM", "LBMJAMES128"]:
            # Lean body weight using James' method with a multiplier of 128 for males
            if self.patient_gender is None:
                logging.warning("Patient gender was not stored in dicom header. LBM cannot be calculated using James\' method.")
                return None
            else:
                if self.patient_gender.lower() in ["m"]:
                    norm_factor = 1.10 * self.patient_weight - 128.0 * (self.patient_weight ** 2.0 / (self.patient_height * 100.0) ** 2.0)

                    # From kg to g
                    norm_factor *= 1000.0

                elif self.patient_gender.lower() in ["f", "w"]:
                    norm_factor = 1.07 * self.patient_weight - 148.0 * (self.patient_weight ** 2.0 / (self.patient_height * 100.0) ** 2.0)

                    # From kg to g
                    norm_factor *= 1000.0
                else:
                    logging.warning("Patient gender was indeterminate. LBM cannot be calculated using James\' method.")
                    return None

            norm_factor = None
        elif suv_type == "LBMJANMA":
            # Lean body weight using the Janmahasatian's method (male : 9,270 × weight/(6,680 + 216 × BMI); female: 9,270 × weight/(8,780 + 244 × BMI))
            # BMI = weight/height^2 (weight in kg, height in m) (10.2165/00003088-200544100-00004)
            if self.patient_gender is None:
                logging.warning("Patient gender was not stored in dicom header. LBM cannot be calculated using Janmahasatian\'s method.")
                return None
            else:
                # Compute bmi
                bmi = self.patient_weight / (self.patient_height**2.0)
                if self.patient_gender.lower() in ["m"]:
                    norm_factor = 9270.0 * self.patient_weight / (6680.0 + 216.0 * bmi)

                    # From kg to g
                    norm_factor *= 1000.0

                elif self.patient_gender.lower() in ["f", "w"]:
                    norm_factor = 9270.0 * self.patient_weight / (8780.0 + 244.0 * bmi)

                    # From kg to g
                    norm_factor *= 1000.0
                else:
                    logging.warning("Patient gender was indeterminate. LBM cannot be calculated using Janmahasatian\'s method.")
                    return None

        elif suv_type == "IBW":
            if self.patient_gender is None:
                logging.warning("Patient gender was not stored in dicom header. IBW cannot be calculated.")
                return None
            else:

                if self.patient_gender.lower() in ["m"]:
                    norm_factor = 48.0 + 1.06 * (self.patient_height * 100.0 - 152.0)

                    # From kg to g
                    norm_factor *= 1000.0

                elif self.patient_gender.lower() in ["f", "w"]:
                    norm_factor = 45.5 + 0.91 * (self.patient_height * 100.0 - 152.0)

                    # From kg to g
                    norm_factor *= 1000.0
                else:
                    logging.warning("Patient gender was indeterminate. LBM cannot be calculated using Janmahasatian\'s method.")
                    return None

        else:
            raise ValueError(f"{suv_type} is not a valid SUV normalisation method.")

        return norm_factor

    def update_dicom_header(self, dcm):

        # Update unit of pixel values
        voxel_unit = "CM2ML" if self.suv_type == "BSA" else "GML"
        set_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1001), value= voxel_unit)

        # Update the SUV type
        set_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1006), value=self.suv_type)

        # Decay correction
        set_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0054, 0x1102), value=self.decay_correction)

        # Add DECY to the image corrections, if this was not done previously.
        image_corrections = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0028, 0x0051), tag_type="mult_str", default=[])
        if "DECY" not in image_corrections:
            image_corrections += ["DECY"]
        set_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0028, 0x0051), value=image_corrections)

        # Update the image type
        image_type = get_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0008, 0x0008), tag_type="mult_str", default=[])
        if len(image_type) > 2:
            image_type[0] = "DERIVED"
            image_type[1] = "SECONDARY"
        else:
            image_type = ["DERIVED", "SECONDARY"]
        set_pydicom_meta_tag(dcm_seq=dcm, tag=(0x0008, 0x0008), value=image_type)

        return dcm

# def suv_list_update(suv_obj_list):
#     """
#     Update settings based on multiple SUV objects (e.g. slices for the same patient)
#     :param suv_obj_list: list of suv class pbjects
#     :return:
#     """
#
#     # Set initial (lowest) acquisition start time for all suv objects
#     overall_start_time = min(get_valid_elements(input_list=[suv_obj.start_ref_time for suv_obj in suv_obj_list]))
#
#     # Get most common radionucleide administration time in case it is missing
#     common_radio_admin_time = get_most_common_element(input_list=get_valid_elements(input_list=[suv_obj.radio_admin_ref_time for suv_obj in suv_obj_list]))
#
#     # Get most common total dose in case it is missing
#     common_total_dose = get_most_common_element(input_list=get_valid_elements(input_list=[suv_obj.total_dose for suv_obj in suv_obj_list]))
#
#     # Get most common radionucleitide half-life in case it is missing from the list
#     common_half_life = get_most_common_element(input_list=get_valid_elements(input_list=[suv_obj.half_life for suv_obj in suv_obj_list]))
#
#     # Get most commonly reported gender
#     common_gender = get_most_common_element(input_list=get_valid_elements(input_list=[suv_obj.patient_gender for suv_obj in suv_obj_list]))
#
#     # Get most commonly reported height
#     common_height = get_most_common_element(input_list=get_valid_elements(input_list=[suv_obj.patient_height for suv_obj in suv_obj_list]))
#
#     # Get most commonly reported weight
#     common_weight = get_most_common_element(input_list=get_valid_elements(input_list=[suv_obj.patient_weight for suv_obj in suv_obj_list]))
#
#     for suv_obj in suv_obj_list:
#         suv_obj.start_ref_time = overall_start_time
#         suv_obj.radio_admin_start_time = common_radio_admin_time
#         suv_obj.total_dose = common_total_dose
#         suv_obj.half_life = common_half_life
#         suv_obj.patient_gender = common_gender
#         suv_obj.patient_height = common_height
#         suv_obj.patient_weight = common_weight
#
#     return suv_obj_list
